# author: rste
# date: 31.08.2020
# (c) AMS AG

#this file provides a class to for using I2C with a digital pattern instrument
# the driver uses a pattern file that was created in NI digital pattern editor and works together with this code
# the pinmap has to be supplied so the pattern can be compiled and run

# author: rste
# date: 31.08.2020
# (c) AMS AG

#this file provides a class to for using I2C with a digital pattern instrument
# the driver uses a pattern file that was created in NI digital pattern editor and works together with this code
# the pinmap has to be supplied so the pattern can be compiled and run

import nidigital
import numpy as np
import logging
import os
import ams_ni_file_tools
from ams_instrument import ams_instrument
from .abstraction import ams_I2C_communication

from typing import List, Tuple


i2cLogger = logging.getLogger(__name__)

class PXI_digital_pattern_I2C(ams_I2C_communication):
    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: I2C communication object
        """
        self.deviceDict = device_dict
        #extract configuration
        self.session = self.deviceDict['Details']['Session']
        if 'speed' in self.deviceDict['Details']:
            self.frequency = self.deviceDict['Details']['speed']
        else:
            i2cLogger.warning("no I2C speed provided, using default: 100kHz")
            self.frequency = 100000
        self.sdaChannel = self.deviceDict['Details']["settings"]['SDA']
        self.sclChannel = self.deviceDict['Details']["settings"]['SCL']
        self.pinmapPath = self.deviceDict['Details']["ni_pin_map_file"]
        if 'patterns_directory' in self.deviceDict['Details']:
            self.patternsDirectory = self.deviceDict['Details']['patterns_directory']
        else:
            i2cLogger.warning(r"no directory for created digital patterns provided, using default: 'C:\NIDIGITALPATTERNS'")
            self.patternsDirectory = r"C:\NIDIGITALPATTERNS"
        #make sure the patterns directory exists
        if not os.path.exists(self.patternsDirectory):
                os.makedirs(self.patternsDirectory)    
        if 'max_write_length' in self.deviceDict['Details']["settings"]:
            self.maxWriteLength = self.deviceDict['Details']["settings"]['max_write_length']
        else:
            i2cLogger.warning(r"no maximum write packet length provided, using default: 10")
            self.maxWriteLength = 10  
        self.busVoltage =   self.deviceDict['Details']["bus_voltage"]
        pass

    def __build_namings(self):
        self.writePatternName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_write"
        self.writeWaveformName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_writeData"
        self.writeDataLoopLabel = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_writeDataLoop"

        self.writeSrcPath = os.path.join(self.patternsDirectory, self.writePatternName + ".digipatsrc")

        self.readPatternName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_read"
        self.readAddressName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_readAddress" # name of the source waveform (for device address)
        self.readWaveformName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_readData"
        self.readDataLoopLabel = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_readDataLoop"
        self.readLastByteLabel = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_readLastByte"

        self.readSrcPath = os.path.join(self.patternsDirectory, self.readPatternName + ".digipatsrc")

        self.registerReadPatternName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_registerRead"
        self.registerReadAddressName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_registerReadAddress" # name of the source waveform (for device and memory address)
        self.registerReadWaveformName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_registerReadData"
        self.registerReadDataLoopLabel = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_registerReadDataLoop"
        self.registerReadLastByteLabel = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_registerReadLastByte"

        self.registerReadSrcPath = os.path.join(self.patternsDirectory, self.registerReadPatternName + ".digipatsrc")

        pass

    def __create_timesets(self):
        i2cLogger.debug("add timings and load them to the instrument (I2C_Bit, I2C_Start, I2C_Stop, I2C_Idle, I2C_SourceStart)")
        
        #timing for a bit
        self.bitTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_Bit"
        self.session.create_time_set(self.bitTimesetName)
        
        #timing for start condition
        self.startTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_Start"
        self.session.create_time_set(self.startTimesetName)
        
        #timing for repeated start condition
        self.repeatedStartTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_RepeatedStart"
        self.session.create_time_set(self.repeatedStartTimesetName)
        
        #timing for stop condition
        self.stopTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_Stop"
        self.session.create_time_set(self.stopTimesetName)
        
        ##timing for i2c bus idle
        self.idleTime = 0.0000013 # 1.3us
        self.idleTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_Idle"
        self.session.create_time_set(self.idleTimesetName)
        
        ##timing for digipat instrument to arm source data
        self.armSourceTime = 0.000003 # 3us
        self.sourceStartTimesetName = "I2C_" + self.sdaChannel + "_" + self.sclChannel + "_SourceStart"
        self.session.create_time_set(self.sourceStartTimesetName)
        
        pass

    def __configure_timesets(self):
        #calculate period
        self.period = 1 / self.frequency

        #timing for a bit
        self.session.configure_time_set_period(self.bitTimesetName, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_drive_edges(self.bitTimesetName, nidigital.DriveFormat.NR, 0.1*self.period , 0.1*self.period, self.period, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_compare_edges_strobe(self.bitTimesetName, 0.5 * self.period)
        self.session.pins[self.sclChannel].configure_time_set_drive_edges(self.bitTimesetName, nidigital.DriveFormat.RH, 0 , 0, 0.5*self.period, 0.5*self.period)
        self.session.pins[self.sclChannel].configure_time_set_compare_edges_strobe(self.bitTimesetName, 0.25 * self.period)
        
        #timing for the start condition
        self.session.configure_time_set_period(self.startTimesetName, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_drive_edges(self.startTimesetName, nidigital.DriveFormat.NR, 0.75*self.period , 0.75*self.period, self.period, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_compare_edges_strobe(self.startTimesetName, 0.5 * self.period)
        self.session.pins[self.sclChannel].configure_time_set_drive_edges(self.startTimesetName, nidigital.DriveFormat.RH, 0 , 0, 0.5*self.period, 0.5*self.period)
        self.session.pins[self.sclChannel].configure_time_set_compare_edges_strobe(self.startTimesetName, 0.25 * self.period)
        
        #timing for the repeated start condition
        self.session.configure_time_set_period(self.repeatedStartTimesetName, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_drive_edges(self.repeatedStartTimesetName, nidigital.DriveFormat.RL, 0, 0, 0.75 * self.period, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_compare_edges_strobe(self.repeatedStartTimesetName, 0.5 * self.period)
        self.session.pins[self.sclChannel].configure_time_set_drive_edges(self.repeatedStartTimesetName, nidigital.DriveFormat.RH, 0 , 0, 0.5*self.period, 0.5*self.period)
        self.session.pins[self.sclChannel].configure_time_set_compare_edges_strobe(self.repeatedStartTimesetName, 0.25 * self.period)
        
        #timing for the stop condition
        self.session.configure_time_set_period(self.stopTimesetName, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_drive_edges(self.stopTimesetName, nidigital.DriveFormat.NR, 0.1*self.period , 0.1*self.period, self.period, self.period)
        self.session.pins[self.sdaChannel].configure_time_set_compare_edges_strobe(self.stopTimesetName, 0.5 * self.period)
        self.session.pins[self.sclChannel].configure_time_set_drive_edges(self.stopTimesetName, nidigital.DriveFormat.RH, 0 , 0, 0.5*self.period, 0.5*self.period)
        self.session.pins[self.sclChannel].configure_time_set_compare_edges_strobe(self.stopTimesetName, 0.25 * self.period)
        
        #timing for bus idle vector
        self.session.configure_time_set_period(self.idleTimesetName, self.idleTime)
        self.session.pins[self.sdaChannel, self.sclChannel].configure_time_set_drive_edges(self.idleTimesetName, nidigital.DriveFormat.NR, 0 , 0, 0.9*self.idleTime, 0.9*self.idleTime)
        self.session.pins[self.sdaChannel, self.sclChannel].configure_time_set_compare_edges_strobe(self.idleTimesetName, 0.5 * self.period)
        
        #timing for the vector machine to arm the source memory
        self.session.configure_time_set_period(self.sourceStartTimesetName, self.armSourceTime)
        self.session.pins[self.sdaChannel, self.sclChannel].configure_time_set_drive_edges(self.sourceStartTimesetName, nidigital.DriveFormat.NR, 0 , 0, 0.9*self.armSourceTime, 0.9*self.armSourceTime)
        self.session.pins[self.sdaChannel, self.sclChannel].configure_time_set_compare_edges_strobe(self.sourceStartTimesetName, 0.5 * self.period)

        pass

    def __set_voltage_levels(self):
        self.session.pins[self.sdaChannel, self.sclChannel].configure_voltage_levels(0, self.busVoltage, 0.3*self.busVoltage, 0.7*self.busVoltage, self.busVoltage)

    def __build_patterns(self, force_recompile = False):
        #build and compile only if there is no pattern already available
        if (not os.path.exists(self.writeSrcPath) ) or force_recompile:
            #build write pattern
            writePattern = ams_ni_file_tools.DigipatSource()
            writePattern.setPatternName(self.writePatternName)
            writePattern.useTimesets(self.idleTimesetName, self.sourceStartTimesetName, self.startTimesetName, self.bitTimesetName, self.stopTimesetName)
            writePattern.useChannels(self.sdaChannel, self.sclChannel)
            writePattern.addVector(self.idleTimesetName, ["X", 1], ["source_d_replace(0, X)"])                          # convert 1 to X (tri state) for sda
            writePattern.addVector(self.idleTimesetName, ["H", 1], ["source_start(" + self.writeWaveformName + ")"])    # SDA must be high at the start, start sourcing
            writePattern.addVector(self.sourceStartTimesetName, ["H", 1])                                               # arm source memory
            writePattern.addVector(self.startTimesetName, [0,1])                                                        # start condition
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address MSB
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address LSB
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # write bit    
            writePattern.addVector(self.bitTimesetName, ["L", 0])                                                       # check for ACK
            writePattern.addVector(self.idleTimesetName, ["X", 1], ["set_loop(reg0)"])                                  # setup the data loop (reg0 must be written via nidigital beforehand!)
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"], self.writeDataLoopLabel)                  # 8 bit data MSB
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            writePattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 8 bit data LSB
            writePattern.addVector(self.bitTimesetName, ["L", 0], ["end_loop("+self.writeDataLoopLabel+")"])            # check for ACK, go back to the start of the data loop if there is a byte left
            writePattern.addVector(self.stopTimesetName, [0,0])                                                         # end of I2C transmission
            writePattern.addVector(self.idleTimesetName, ["X", 1], ["halt"])                                                      # halt the vector sequencer and put the pins in idle state

            writePattern.generateDigipatsrcFile(self.writeSrcPath)

            #compile write pattern
            writeCompileOk, self.writePatPath = ams_ni_file_tools.compileDigipatSource(self.writeSrcPath, self.pinmapPath)

        else:
            writeCompileOk = True
            self.writePatPath = self.writeSrcPath[:-3]

        #build and compile only if there is no pattern already available
        if ( not os.path.exists(self.readSrcPath) ) or force_recompile:
            #build read pattern
            readPattern = ams_ni_file_tools.DigipatSource()
            readPattern.setPatternName(self.readPatternName)
            readPattern.useTimesets(self.idleTimesetName, self.sourceStartTimesetName, self.startTimesetName, self.bitTimesetName, self.stopTimesetName)
            readPattern.useChannels(self.sdaChannel, self.sclChannel)
            readPattern.addVector(self.idleTimesetName, ["X", 1], ["source_d_replace(0, X)"])                          # convert 1 to X (tri state) for sda
            readPattern.addVector(self.idleTimesetName, ["H", 1], ["source_start(" + self.readAddressName + ")"])      # SDA must be high at the start, start sourcing the address
            readPattern.addVector(self.idleTimesetName, ["H", 1], ["capture_start(" + self.readWaveformName + ")"])    # start capturing into read data waveform
            readPattern.addVector(self.sourceStartTimesetName, ["H", 1])                                               # arm source memory
            readPattern.addVector(self.startTimesetName, [0,1])                                                        # start condition
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address MSB
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address LSB
            readPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # write bit    
            readPattern.addVector(self.bitTimesetName, ["L", 0] )                                           # check for ACK
            readPattern.addVector(self.idleTimesetName, ["X", 1], ["jump_if(seqflag0, "+self.readLastByteLabel+")"])   # jump over the data loop if there is only one byte to be read (set seqflag0 beforehand in this case!)
            readPattern.addVector(self.idleTimesetName, ["X", 1], ["set_loop(reg0)"])                                  # setup data loop (reg0 must be written before bursting!)
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"], self.readDataLoopLabel)                                          # 8 bit data MSB
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])                                          # 8 bit data LSB
            readPattern.addVector(self.bitTimesetName, [0, 0], ["end_loop(" + self.readDataLoopLabel + ")"])           # give ACK
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"], self.readLastByteLabel)                  # 8 bit data MSB
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            readPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])                                          # 8 bit data LSB
            readPattern.addVector(self.bitTimesetName, ["X", 0])                                                       # give NACK
            readPattern.addVector(self.stopTimesetName, [0,0], ["capture_stop"])                                                         # end of I2C transmission
            readPattern.addVector(self.idleTimesetName, ["X", 1], ["halt"])                                                    # halt the vector sequencer and put the pins in idle state
       
            readPattern.generateDigipatsrcFile(self.readSrcPath)

            #compile read pattern
            readCompileOk, self.readPatPath = ams_ni_file_tools.compileDigipatSource(self.readSrcPath, self.pinmapPath)
        else:
            readCompileOk = True
            self.readPatPath = self.readSrcPath[:-3]


        #build and compile only if there is no pattern already available
        if ( not os.path.exists(self.registerReadSrcPath) ) or force_recompile:
            #build register read pattern
            registerReadPattern = ams_ni_file_tools.DigipatSource()
            registerReadPattern.setPatternName(self.registerReadPatternName)
            registerReadPattern.useTimesets(self.idleTimesetName, self.sourceStartTimesetName, self.startTimesetName, self.bitTimesetName, self.stopTimesetName, self.repeatedStartTimesetName)
            registerReadPattern.useChannels(self.sdaChannel, self.sclChannel)
            registerReadPattern.addVector(self.idleTimesetName, ["X", 1], ["source_d_replace(0, X)"])                          # convert 1 to X (tri state) for sda
            registerReadPattern.addVector(self.idleTimesetName, ["H", 1], ["source_start(" + self.registerReadAddressName + ")"])      # SDA must be high at the start, start sourcing the address
            registerReadPattern.addVector(self.idleTimesetName, ["H", 1], ["capture_start(" + self.registerReadWaveformName + ")"])    # start capturing into read data waveform
            registerReadPattern.addVector(self.sourceStartTimesetName, ["H", 1])                                               # arm source memory
            registerReadPattern.addVector(self.startTimesetName, [0,1] )                                                        # start condition - write address
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address MSB
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address LSB
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # write bit    
            registerReadPattern.addVector(self.bitTimesetName, ["L", 0] )                                           # check for ACK
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 8 bit register address MSB
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 8 bit register address LSB    
            registerReadPattern.addVector(self.bitTimesetName, ["L", 0] )                                           # check for ACK
            #registerReadPattern.addVector(self.stopTimesetName, [1,1])                                                         # stop condition before issuing next command
            registerReadPattern.addVector(self.repeatedStartTimesetName, [1,0])                                                        # repeated start condition - read data
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address MSB
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"]) 
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # 7 bit address LSB
            registerReadPattern.addVector(self.bitTimesetName, ["D", 0], ["source"])                                           # read bit    
            registerReadPattern.addVector(self.bitTimesetName, ["L", 0] )                                           # check for ACK
            registerReadPattern.addVector(self.idleTimesetName, ["X", 1], ["jump_if(seqflag0, "+self.registerReadLastByteLabel+")"])   # jump over the data loop if there is only one byte to be read (set seqflag0 beforehand in this case!)
            registerReadPattern.addVector(self.idleTimesetName, ["X", 1], ["set_loop(reg0)"])                                  # setup data loop (reg0 must be written before bursting!)
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"], self.registerReadDataLoopLabel)                                          # 8 bit data MSB
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])                                          # 8 bit data LSB
            registerReadPattern.addVector(self.bitTimesetName, [0, 0], ["end_loop(" + self.registerReadDataLoopLabel + ")"])           # give ACK
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"], self.registerReadLastByteLabel)                  # 8 bit data MSB
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])
            registerReadPattern.addVector(self.bitTimesetName, ["V", 0], ["capture"])                                          # 8 bit data LSB
            registerReadPattern.addVector(self.bitTimesetName, ["X", 0])                                                       # give NACK
            registerReadPattern.addVector(self.stopTimesetName, [0,0], ["capture_stop"])                                                         # end of I2C transmission
            registerReadPattern.addVector(self.idleTimesetName, ["X", 1], ["halt"])                                                    # halt the vector sequencer and put the pins in idle state

            registerReadPattern.generateDigipatsrcFile(self.registerReadSrcPath)

            #compile register read pattern
            registerReadCompileOk, self.registerReadPatPath = ams_ni_file_tools.compileDigipatSource(self.registerReadSrcPath, self.pinmapPath)
        else:
            registerReadCompileOk = True
            self.registerReadPatPath = self.registerReadSrcPath[:-3]

        return (writeCompileOk and readCompileOk and registerReadCompileOk)

    def __load_patterns(self):
        #load write pattern in the instrument
        self.session.load_pattern(self.writePatPath)
        #load register read pattern in the instrument
        self.session.load_pattern(self.registerReadPatPath)
        #load read pattern in the instrument
        self.session.load_pattern(self.readPatPath)
        
    def I2C_init(self, device_name: str = "") -> bool:
        """
        This method initializes a I2C master on a digital pattern instrument

        Args:
            device_name ([type]): [description]

        Returns:
            Bool: True if the Initialization worked, False otherwise
        """       

            

        self.__build_namings()
        
        #(edit pinmap) the channels specified for SDA and SCL must be defined in the pinmap beforehand!

        #add timing for i2c communications
        self.__create_timesets()
        self.__configure_timesets()

        #set the voltage levels
        self.__set_voltage_levels()
        
        patternsOk = self.__build_patterns()

        self.__load_patterns()

        return patternsOk

    def I2C_configure(
        self,
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
        ) -> bool:
        """I2C_configure 
        This method configures the communication

        Args:
            speed (int): communication speed (bps)
            bus_voltage (float): bus voltage in V
            settings_dict (dict, optional): defines any implementation setting that may be necessary. Defaults to {}.
        Returns:
            bool: True on success
        """

        if not self.frequency == speed:
            self.frequency = speed
            self.__configure_timesets()

        if not self.busVoltage == bus_voltage:
            self.busVoltage = bus_voltage
            self.__set_voltage_levels()
        if settings_dict:
            recompile = False
            if 'SDA' in settings_dict:
                if not (self.sdaChannel == settings_dict['SDA']):
                    self.sdaChannel = settings_dict['SDA']
                    recompile = True

            if 'SCL' in settings_dict:
                if not (self.sclChannel == settings_dict['SCL']):    
                    self.sclChannel = settings_dict['SCL']
                    recompile = True
                    

            if recompile:
                self.__build_namings()
                self.__build_patterns()
                self.__load_patterns()
        pass

    def I2C_write(
        self, 
        address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_write writes the Bytes in payload to an IC at address

        Args:
            address (int): 8 bit address of the IC (7 bit address << 1)
            payload (List[int]): A list of the payload bytes to be transmitted

        Returns:
            bool: True on success, False if the Slave doesn't respond
        """

        i2cLogger.debug("writing data '" + str([hex(i) for i in payload]) + "' to address: '" + hex(address) + "'")
        #build source waveform
        i2cLogger.debug("building address + data waveform ...")
        self.session.channels[self.sdaChannel].create_source_waveform_serial(self.writeWaveformName, nidigital.SourceDataMapping.BROADCAST, 8, nidigital.BitOrder.MSB)

        deviceAddress = address
        #ensure write bit in the address
        deviceAddress = deviceAddress & 0xFE
        #add all data
        theDataToWrite = [deviceAddress]
        for datum in payload:
            theDataToWrite.append(datum)

        #fill the remaining void with 0. The source waveform must always be written in the same format
        while len(theDataToWrite) < self.maxWriteLength:
            theDataToWrite.append(0)
        
        i2cLogger.debug("i2c writedata waveform name: " + self.writeWaveformName)
        i2cLogger.debug("full I2C_writeData waveform: " + str([hex(i) for i in theDataToWrite]))
        #load data waveform into the instrument
        self.session.write_source_waveform_broadcast(self.writeWaveformName, theDataToWrite)

        #run pattern
        #determine how many bytes to write
        read_count = len(payload)
        i2cLogger.debug("running pattern with " + str(read_count) + " Bytes of data")
        #provide the digital pattern sequencer with this information on register 0 (used by the pattern internally to check the number of bytes)
        self.session.write_sequencer_register(nidigital.SequencerRegister.REGISTER0, read_count)
        self.session.burst_pattern(self.writePatternName)

        #check pattern pass/fail to determine if slave was responding
        return self.session.get_site_pass_fail()[0]

    def I2C_read(
        self, 
        address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_read 
           read read_count bytes from a slave at address

        Args:
            address (int): 8 bit address of the slave (7 bit address << 1)
            read_count (int): number of bytes to read

        Returns:
            Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """


        i2cLogger.debug("reading data from address: '" + hex(address) + "'")

        #build source waveform
        i2cLogger.debug("building address waveform ...")
        self.session.channels[self.sdaChannel].create_source_waveform_serial(self.readAddressName, nidigital.SourceDataMapping.BROADCAST, 8, nidigital.BitOrder.MSB)
        deviceAddress = address
        #add read bit to the address 
        deviceAddress = deviceAddress | 0x01
        #add all data
        theDataToWrite = [deviceAddress]
        #load data waveform into the instrument
        self.session.write_source_waveform_broadcast(self.readAddressName, theDataToWrite)

        #setup capture waveform
        i2cLogger.debug("setup capture waveform")
        self.session.channels[self.sdaChannel].create_capture_waveform_serial(self.readWaveformName, 8, nidigital.BitOrder.MSB)

        #run pattern
        #determine how many bytes to write
        i2cLogger.debug("running pattern with " + str(read_count) + " Bytes of data")
        #the pattern sequencer needs the number of bytes to read minus 1 on seq0 register
        self.session.write_sequencer_register(nidigital.SequencerRegister.REGISTER0, read_count-1)
        #additionally, the pattern needs the information if only one byte is read in seqflag0
        if (read_count == 1):
            self.session.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, True)
        else:
            self.session.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, False)
        
        self.session.burst_pattern(self.readPatternName)

        #retreive Data from the captured waveform
        capturedWaveformMemoryViewPerSite = self.session.fetch_capture_waveform(self.readWaveformName,read_count)

        capturedWaveformBytes = capturedWaveformMemoryViewPerSite[0].tobytes()
        capturedWaveformBytes = capturedWaveformBytes[0::4] #theBytes of the memory view are contacenated uint32, but only the most significant byte contains data
        capturedWaveformBytes = [int(byte) for byte in capturedWaveformBytes] #convert from bytes to int

        i2cLogger.debug("read " + str(read_count) + " Bytes of data from address " + hex(address) + ": " + str([hex(i) for i in capturedWaveformBytes]) )

        #check pattern pass/fail to determine if slave was responding
        patternPassFail = self.session.get_site_pass_fail()[0]

        return patternPassFail, capturedWaveformBytes

    def I2C_register_write(
        self, 
        device_address: int, 
        register_address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_register_write 
        Write bytes to the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to write to
            payload (List[int]): List of bytes to write to the device

        Returns:
            bool: True on success
        """

        return self.I2C_write(device_address, [register_address] + payload)

    def I2C_register_read(
        self, 
        device_address: int, 
        register_address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_register_read 
        Read bytes from the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to read
            read_count (int): Number of bytes to read

        Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """
        i2cLogger.debug("reading data from address: '" + hex(device_address) + "' as atomic operation!")

        #build source waveform
        i2cLogger.debug("building address waveform ...")
        self.session.channels[self.sdaChannel].create_source_waveform_serial(self.registerReadAddressName, nidigital.SourceDataMapping.BROADCAST, 8, nidigital.BitOrder.MSB)
        #add all data
        theDataToWrite = [device_address & 0xFE, register_address, device_address | 0x01]

        #load data waveform into the instrument
        self.session.write_source_waveform_broadcast(self.registerReadAddressName, theDataToWrite)

        #setup capture waveform
        i2cLogger.debug("setup capture waveform")
        self.session.channels[self.sdaChannel].create_capture_waveform_serial(self.registerReadWaveformName, 8, nidigital.BitOrder.MSB)

        #run pattern
        #determine how many bytes to read
        i2cLogger.debug("running pattern with " + str(read_count) + " Bytes of data")
        #the pattern sequencer needs the number of bytes to read minus 1 on seq0 register
        self.session.write_sequencer_register(nidigital.SequencerRegister.REGISTER0, read_count-1)
        #additionally, the pattern needs the information if only one byte is read in seqflag0
        if (read_count == 1):
            self.session.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, True)
        else:
            self.session.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, False)
        
        self.session.burst_pattern(self.registerReadPatternName)

        #retreive Data from the captured waveform
        capturedWaveformMemoryViewPerSite = self.session.fetch_capture_waveform(self.registerReadWaveformName,read_count)

        capturedWaveformBytes = capturedWaveformMemoryViewPerSite[0].tobytes()
        capturedWaveformBytes = capturedWaveformBytes[0::4] #theBytes of the memory view are contacenated uint32, but only the most significant byte contains data
        capturedWaveformBytes = [int(byte) for byte in capturedWaveformBytes] #convert from bytes to int

        i2cLogger.debug("read " + str(read_count) + " Bytes of data from device " + hex(device_address) + ", register " + hex(register_address) + ": " + str([hex(i) for i in capturedWaveformBytes]) )

        #check pattern pass/fail to determine if slave was responding
        patternPassFail = self.session.get_site_pass_fail()[0]

        return patternPassFail, capturedWaveformBytes


    def GetInstrumentError(self):
        return " not implemented!"

    def IdentifyInstrument(self):
        return " AMS I2C Master on PXIe Digital Pattern Instrument "

    def ResetInstrument(self):
        return " not implemented! Reset Digital Pattern Instrument instead!"
