import logging
from typing import List, Tuple
import nidigital
import os

import ams_ni_file_tools


from .abstraction import ams_SPI_communication

logger = logging.getLogger(__name__)


"""

{
    "Instrument_name": "SPI_Master_1",
    "Type": "ams_communication.PXI_digital_pattern_SPI",
    "Details": {
        "Session": the nidigital session,
        "ni_pin_map_file": "ni_pinmap_file.pinmap",
        "patterns_directory": "\\spipatterns\\",
        "speed": 100000,
        "bus_voltage": 5,
        "settings": {
            "MISO"  : "CBIT_I2C_SCL",
            "MOSI"  : "CBIT_I2C_SDA",
            "SCLK"  : "",
            "CS"    : "",
            "CPOL"  : 1,
            "CPHA"  : 0,
            "bit_order": "MSB"
        }
    }
}

"""

class PXI_digital_pattern_SPI(ams_SPI_communication):
    """
    This implementation enables SPI to be used on a digital pattern instrument
    """

    SPI_CHANNELS = ["MISO", "MOSI", "SCLK", "CS"]

    #pattern template to fill in timesets and channels
    TEMPLATE = """
file_format_version 1.1;
timeset $no_clk_bit, $clk_bit, $client_select_delay, $client_deselect_delay, $wait;

pattern $pattern_name ($mosi_channel, $miso_channel, $sclk_channel, $cs_channel)
{
source_start($source_wfm)	            $no_clk_bit	                -   -   -   -; //start sourcing
	                                    $wait 	                    -   -   -   -; //wait for source memory to arm
capture_start($capture_wfm)	            $no_clk_bit	                -   -   -   -; //start capturing
                                        $no_clk_bit                 0   X   $sclk_idl   1; //level state before selecting client
set_loop(reg0)                          $client_select_delay        0   X   $sclk_idl   0; //select client
spi_loop:   source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 0
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 1
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 3
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 2
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 4
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 5
            source, capture             $clk_bit                    D   V   $sclk_clk   0; //bit 6
end_loop(spi_loop), source, capture     $clk_bit                    D   V   $sclk_clk   0; //bit 7
                                        $client_deselect_delay      0   X   $sclk_idl   0; //wait after the full bytes are transfered
                                        $no_clk_bit                 0   X   $sclk_idl   1; //deselect client
capture_stop	                        $no_clk_bit	                -	-   -   -; //stop capturing
halt	                                $no_clk_bit	                -	-   -   -; //end of pattern
}
"""



    sclk_timing_edges = [{  "format": nidigital.DriveFormat.SBC, #CPHA = 0
                            "drive_on_edge": 0,
                            "drive_data_edge": 0.5,
                            "drive_return_edge": 1,
                            "drive_off_edge": 1,
                            "strobe_edge": 0.75     },
                         {  "format": nidigital.DriveFormat.SBC, #CPHA = 1
                            "drive_on_edge": 0,
                            "drive_data_edge": 0,
                            "drive_return_edge": 0.5,
                            "drive_off_edge": 1,
                            "strobe_edge": 0.25 }]

    def _get_driver_id(self) -> str:
        """returns a unique id for this driver

        Returns:
            str: the id
        """
        id = ""
        for channel in self.SPI_CHANNELS:
            id = id + self._device_dict['Details']['settings'][channel] + "_"
        return id[:-1]

    @staticmethod
    def _fill_default_dict_parameter(dictionary:dict, parameter:str, default:object) -> bool:
        """_fill_dict_parameter

            sets a dict parameter to the default parameter if it is not present in the dict
            also logs a warning when no value is given

        Args:
            dictionary (dict): The dictionary to fill the default parameter
            parameter (str): The key of the parameter
            default (object): The default value to use for this parameter

        Returns:
            bool: True when there was a value present, False when the default was set
        """

        if not parameter in dictionary:
            logger.warning(f"parameter {parameter} not provided, using default: '{default}'")
            dictionary[parameter] = default
            return False
        
        return True

    def _build_naming_dict(self):
        """generate all namings for waveforms and patterns from the configured driver
        """
        pattern_name = "SPI_" + self._get_driver_id()
        self._naming_dict["pattern_name"] = pattern_name
        self._naming_dict["source_wfm"]   = pattern_name + "_source"
        self._naming_dict["capture_wfm"]  = pattern_name + "_capture"
        self._naming_dict["pattern_path"] = os.path.join(self._device_dict['Details']['patterns_directory'], pattern_name + ".digipat")

    def _build_timing_names(self):
        """generate all namings for timings from the configured driver
        """
        for timeset in self._timing_dict:
            self._timing_dict[timeset]["name"] = "SPI_" + self._get_driver_id() + "_" + timeset

    def _create_timesets(self):
        """_create_timesets create all timesets needed by the spi driver
        """
        for timeset in self._timing_dict:
            self._device_dict['Details']['Session'].create_time_set(self._timing_dict[timeset]["name"])
 
    def _configure_timesets(self):
        """_configure_timesets Set all edges of the timeset depending on the before-set configuration
        """
        period = 1/self._device_dict['Details']['speed']

        #adapt clock edges to the CPHA clock phase
        self._timing_dict["clk_bit"]["edges"]["SCLK"] = self.sclk_timing_edges[self._device_dict["Details"]["settings"]["CPHA"]]

        for timeset in self._timing_dict:
            self._device_dict['Details']['Session'].configure_time_set_period(self._timing_dict[timeset]["name"], self._timing_dict[timeset]["period"] * period)
            for channel in self._timing_dict[timeset]["edges"]:
                self._device_dict['Details']['Session'].pins[self._device_dict['Details']["settings"][channel]]\
                    .configure_time_set_drive_edges(self._timing_dict[timeset]["name"],\
                                                    self._timing_dict[timeset]["edges"][channel]["format"],\
                                                    self._timing_dict[timeset]["edges"][channel]["drive_on_edge"] * period,\
                                                    self._timing_dict[timeset]["edges"][channel]["drive_data_edge"] * period,\
                                                    self._timing_dict[timeset]["edges"][channel]["drive_return_edge"] * period,\
                                                    self._timing_dict[timeset]["edges"][channel]["drive_off_edge"] * period)
                self._device_dict['Details']['Session'].pins[self._device_dict['Details']["settings"][channel]]\
                    .configure_time_set_compare_edges_strobe(self._timing_dict[timeset]["name"],\
                                                             self._timing_dict[timeset]["edges"][channel]["strobe_edge"] * period)

    def _set_voltage_levels(self):
        """_set_voltage_levels Set The voltage levels depending on the before-set configuration
        """
        v_il_ratio = 0
        v_ih_ratio = 1
        v_ol_ratio = 0.3
        v_oh_ratio = 0.7
        v_term_ratio = 1
        self._device_dict['Details']['Session']\
            .pins[ [self._device_dict['Details']["settings"][channel] for channel in self.SPI_CHANNELS ] ]\
            .configure_voltage_levels(  v_il_ratio * self._device_dict['Details']["bus_voltage"],\
                                        v_ih_ratio * self._device_dict['Details']["bus_voltage"],\
                                        v_ol_ratio * self._device_dict['Details']["bus_voltage"],\
                                        v_oh_ratio * self._device_dict['Details']["bus_voltage"],\
                                        v_term_ratio * self._device_dict['Details']["bus_voltage"]  )
        pass

    def _create_waveforms(self) -> bool:
        """create capture and source waveform in the instrument

        Returns:
            bool: True on success
        """
        if self._device_dict['Details']['settings']['bit_order'] == "MSB":
            bit_order = nidigital.BitOrder.MSB
        elif self._device_dict['Details']['settings']['bit_order'] == "LSB":
            bit_order = nidigital.BitOrder.LSB
        else:
            logger.error(f"could not set waveform bit order '{self._device_dict['Details']['settings']['bit_order']}'! Must be 'MSB' or 'LSB'")
            return False
        self._device_dict['Details']['Session'].channels[self._device_dict['Details']["settings"]['MISO']].create_capture_waveform_serial(self._naming_dict["capture_wfm"], 8, bit_order)
        self._device_dict['Details']['Session'].channels[self._device_dict['Details']["settings"]['MOSI']].create_source_waveform_serial(self._naming_dict["source_wfm"], nidigital.SourceDataMapping.BROADCAST, 8, bit_order)
        return True

    def _fill_pattern_template(self, template:str, namings: dict) -> str: 
        """_fill_pattern_template Fills a pattern template with corresponding signal names, timing names, etc

        Args:
            template (str): the template pattern string to fill
            namings (dict): a dict of the namings to use in this pattern

        Returns:
            str: the .digipatsrc pattern filled with the needed signals
        """

        filled = template
        for timeset in self._timing_dict:
            filled = filled.replace("$"+timeset, self._timing_dict[timeset]["name"])
        filled = filled.replace("$pattern_name", namings["pattern_name"])
        filled = filled.replace("$mosi_channel", self._device_dict['Details']["settings"]['MOSI'])
        filled = filled.replace("$miso_channel", self._device_dict['Details']["settings"]['MISO'])
        filled = filled.replace("$sclk_channel", self._device_dict['Details']["settings"]['SCLK'])
        filled = filled.replace("$cs_channel", self._device_dict['Details']["settings"]['CS'])
        filled = filled.replace("$source_wfm", namings["source_wfm"])
        filled = filled.replace("$capture_wfm", namings["capture_wfm"])
        filled = filled.replace("$sclk_idl", ("0","1")[self._device_dict['Details']["settings"]['CPOL']])
        filled = filled.replace("$sclk_clk", ("1","0")[self._device_dict['Details']["settings"]['CPOL']])
        return filled

    def _compile_pattern(self):
        """ generate and compile th spi pattern
        """
      
        pattern = self._fill_pattern_template(self.TEMPLATE, self._naming_dict)
        
        # write the pattern to a file
        with open(self._naming_dict["pattern_path"] + "src", "w") as file:
            file.write(pattern)
        # compile the pattern
        ams_ni_file_tools.compileDigipatSource(self._naming_dict["pattern_path"] + "src", self._device_dict['Details']["ni_pin_map_file"])

    def _load_pattern(self, pattern_path:str):
        """load a pattern from a given path into the digital pattern instrument

        Args:
            pattern_path (str): path to the compiled pattern file
        """
        self._device_dict['Details']['Session'].load_pattern(pattern_path)

    def _transcieve(self, no_bytes:int) -> bool:
        """runs the spi pattern

        Args:
            no_bytes (int): The number of bytes in the communication packet.

        Returns:
            bool: True if the pattern passes, False if it fails
        """

        self._device_dict["Details"]["Session"].write_sequencer_register(nidigital.SequencerRegister.REGISTER0, no_bytes)
        pattern_passfail = self._device_dict['Details']['Session'].burst_pattern(self._naming_dict["pattern_name"], wait_until_done=True)[0]
        return pattern_passfail

    def _fill_source_memory(self, the_bytes: List[int]):
        """Fill the source memory of the instrument

        Args:
            the_bytes (list[int]): a list of integers representing the bytes to send
        """
        self._device_dict['Details']['Session'].write_source_waveform_broadcast(self._naming_dict["source_wfm"], the_bytes)
    
    def _read_capture_memory(self, no_bytes:int) -> List[int]:
        """read the captured memory from the instrument

        Args:
            no_bytes (int): the number of bytes to read from capture memory

        Returns:
            List[int]: a list of integers representing the read bytes
        """

        #fetch the capture
        captured_waveform_memory_view_per_site = self._device_dict['Details']['Session'].fetch_capture_waveform(self._naming_dict["capture_wfm"], no_bytes)
        captured_waveform_bytes = captured_waveform_memory_view_per_site[0]

        logger.debug("captured: "+captured_waveform_bytes.hex())

        return list(captured_waveform_bytes)

    @staticmethod
    def _hex_to_int_list(the_hex:str) -> List[int]:
        """convert a hexadecimal string to a list of integers representing single bytes

        Args:
            the_hex (str): the hexadecimal string

        Returns:
            List[int]: the list of integers
        """
        if (the_hex[:2].upper() == '0X'): #skip possible prefix
            the_hex = the_hex[2:]
        int_list = []
        while len(the_hex) > 1:
            int_list.append( int(the_hex[:2], 16) )
            the_hex = the_hex[2:]
        if len(the_hex) == 1:
            int_list.append( int(the_hex[:2], 16) * 16)

        return int_list

    @staticmethod
    def _int_list_to_hex(the_int_list: List[int]) -> str:
        """convert a list of integers representing single bytes to a hexadecimal string

        Args:
            the_int_list (List[int]): The list of integers

        Returns:
            str: the hexadecimal string
        """
        the_hex = "0x"
        for integer in the_int_list:
            the_hex += f"{integer:02x}"
        return the_hex

    @staticmethod
    def _resize_list_to_length(the_list: List, length:int, padding_value:object = None) -> List:
        """ pad/cut a list to a specific length.
            if the list is longer than the given length, it is cut
            With a shorter list, 'padding_value' objects are appended until the list is at the right length

        Args:
            the_list (List): The list to resize
            length (int): the resized list length
            padding_value (object, optional): The object used for padding. Defaults to None.

        Returns:
            List: The resized list
        """
        if (len(the_list) < length):
            the_list.append([padding_value for i in range(length - len(the_list))])
        
        return the_list[:length]

    @staticmethod
    def _condition_payload(payload:str, length:int) -> List[int]:
        """generate a clean usable integer array with the right length representing the payload

        Args:
            payload (str): The payload as hex string
            length (int): The desired length of the payload

        Returns:
            List[int]: The payload as list of 8 bit integers
        """
        #build up payload
        if not payload:
            payload = []
        else:
            payload = PXI_digital_pattern_SPI._hex_to_int_list(payload)
        #determine payload length
        if not length:
            length = len(payload)    
        #resize payload
        payload = PXI_digital_pattern_SPI._resize_list_to_length(payload, length, 0)

        return payload

    def __init__(self, device_dict: dict) -> object:
        
        self._device_dict = device_dict
        self._fill_default_dict_parameter(self._device_dict['Details'], 'speed', 100000)
        self._fill_default_dict_parameter(self._device_dict['Details'], 'patterns_directory', 'C:\\NIDIGITALPATTERNS')
        #make sure the patterns directory exists
        if not os.path.exists(self._device_dict['Details']['patterns_directory']):
            os.makedirs(self._device_dict['Details']['patterns_directory'])
        self._fill_default_dict_parameter(self._device_dict['Details']['settings'], 'CPOL', 1)
        self._fill_default_dict_parameter(self._device_dict['Details']['settings'], 'CPHA', 0)
        self._fill_default_dict_parameter(self._device_dict['Details']['settings'], 'bit_order', "MSB")
        
        

        self._timing_dict = {
            "wait":{        "name":None, 
                            "period": 0.2,  
                            "edges":{       "MISO" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "MOSI" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "SCLK" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "CS"   :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},

            "clk_bit":{     "name":None, 
                            "period": 1.0,  
                            "edges":{       "MISO" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "MOSI" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "SCLK" :{   "format": nidigital.DriveFormat.RL,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 0.5,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.25     },
                                            "CS"   :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},

            "no_clk_bit":{  "name":None, 
                            "period": 1.0,  
                            "edges":{       "MISO" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "MOSI" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "SCLK" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "CS"   :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},

            "client_select_delay" :{
                            "name":None, 
                            "period": 1.0,  
                            "edges":{       "MISO" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "MOSI" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "SCLK" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "CS"   :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},
            "client_deselect_delay" :{
                            "name":None, 
                            "period": 1.0,  
                            "edges":{       "MISO" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "MOSI" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "SCLK" :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "CS"   :{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},
        }
        self._naming_dict = {}
    
    def SPI_init(self) -> bool:
        """SPI_init
        This Method initializes the SPI driver
        
        Returns:
            Bool: True if the Initialization worked, False otherwise
        """
        #add naming and timing for spi communications
        self._build_naming_dict()
        self._create_waveforms()
        self._build_timing_names()
        self._create_timesets()
        self._configure_timesets()

        #set the voltage levels
        self._set_voltage_levels()

        #compile and load pattern
        if not os.path.exists(self._naming_dict["pattern_path"]):
            self._compile_pattern()

        self._load_pattern(self._naming_dict["pattern_path"])

        return True    

    def SPI_configure(
        self, 
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
        ):
        """SPI_configure 
        This method configures the SPI communication.
        If this method is not called the instance uses the defaults given in the HW config file

        Args:
            speed (int): speed of the communication
            bus_voltage (float): bus voltage level in V
            settings_dict (dict): defines any setting that may be implementation specific as interface pins in case of PXI pattern instrument. Defaults to {}.
        """
        if not self._device_dict['Details']['speed'] == speed:
            self._device_dict['Details']['speed'] = speed
            self._configure_timesets()

        if not self._device_dict['Details']["bus_voltage"] == bus_voltage:
            self._device_dict['Details']["bus_voltage"] = bus_voltage
            self._set_voltage_levels()

        if settings_dict:
            recompile = False
            for channel in self.SPI_CHANNELS:
                if channel in settings_dict:
                    if not (self._device_dict['Details']["settings"][channel] == settings_dict[channel]):
                        self._device_dict['Details']["settings"][channel] = settings_dict[channel]
                        recompile = True

            for mode_bit in ["CPOL", "CPHA"]:
                if mode_bit in settings_dict:
                    recompile = True
                    self._device_dict['Details']['settings'][mode_bit] = settings_dict[mode_bit]

            if recompile:
                self._build_timing_names()
                self._create_timesets()
                self._configure_timesets()
                self._build_naming_dict()
                self._compile_pattern()
                self._create_waveforms()
                self._load_pattern()
        pass

    def SPI_transfer(self, payload:str = None, length:int = None) -> Tuple[bool, str]:
        """SPI_transfer
        This method executes the bidirectional data transfer defined by the SPI protocol.
        I can take either the payload and/or the payload length as parameter
        when both are given, the payload is transmitted up to the length. Too short payloads are padded with 0x00
        when no payload length is given, as many bytes as there are in the payload data are transmitted.
        when no payload is given, 0x00 is transmitted for every byte

        Args:
            payload (str, optional): The payload in hexadecimal format. Defaults to "".
            length (int, optional): The number of bytes to be transmitted. Defaults to None.

        Returns:
            Tuple[bool, str]: [True on success, The read payload as hexadecimal string]
        """
        
        #setup transmitted payload
        payload = self._condition_payload(payload, length)
        no_bytes = len(payload)
        self._fill_source_memory(payload)

        logger.debug(f"transfering spi payload:{payload}")

        #execute
        pattern_passfail = self._transcieve(no_bytes)

        #read received payload
        read_data = self._int_list_to_hex(self._read_capture_memory(no_bytes))

        return (pattern_passfail, read_data)
