
import logging
import nidigital
import time

import os
import math

from bitstream import BitStream
from numpy import uint8
from numpy import uint16
from numpy import uint32
import numpy as np
from typing import Dict, Tuple, List

swdLogger = logging.getLogger(__name__)

import ams_ni_file_tools
from .abstraction import ams_SWD_communication

#header pattern needed by every pattern
TEMPLATE_HEADER = """
file_format_version 1.1;
timeset $no_clk_bit, $clk_bit, $wait, $wait_short;

pattern $pattern_name ($swclk_channel, $swdio_channel)
{
source_start($source_wfm)	    $no_clk_bit	    -   -; //start sourcing
	                            $wait 	        -   -; //wait for source memory to arm
capture_start($capture_wfm)	    $no_clk_bit	    -   -; //start capturing
"""
template_header = { "digipatsrc"    : TEMPLATE_HEADER,
                    "source_bits"   : 0,
                    "capture_bits"  : 0,
                    "identifier"    : 'hdr' }

TEMPLATE_OUTPUT_NIBBLE = """
repeat(4), source   $clk_bit    1   D; //output nibble
"""
template_output_nibble = {   "digipatsrc"    : TEMPLATE_OUTPUT_NIBBLE,
                                    "source_bits"   : 4,
                                    "capture_bits"  : 0,
                                    "identifier"    : 'nib' }

TEMPLATE_OUTPUT_BYTE = """
repeat(8), source   $clk_bit    0   D; //output byte
"""
template_output_byte = {   "digipatsrc"    : TEMPLATE_OUTPUT_BYTE,
                                    "source_bits"   : 8,
                                    "capture_bits"  : 0,
                                    "identifier"    : '' }

#put the 0x7b9e (lsb first) switch command into the waveform!
#16 bits source, 0 bits capture
TEMPLATE_SWITCH_TO_SWD = """
                        $no_clk_bit     0   1;
repeat(100)	            $clk_bit	    1	1; //line reset
repeat(16), source	    $clk_bit	    1	D; //switch to swd
repeat(100)	            $clk_bit	    1	1; //line reset
"""
template_switch_to_swd = {  "digipatsrc"    : TEMPLATE_SWITCH_TO_SWD,
                            "source_bits"   : 16,
                            "capture_bits"  : 0,
                            "identifier"    : 'stswd' }

TEMPLATE_IDLE = """
repeat(8)	            $clk_bit	    0	0; //Idle phase
"""
template_idle = {   "digipatsrc"    : TEMPLATE_IDLE,
                    "source_bits"   : 0,
                    "capture_bits"  : 0,
                    "identifier"    : "i"}

LINE_RESET50 = """
repeat(50)	            $clk_bit	    0	1; //Idle phase
"""
line_reset50 = {   "digipatsrc"    : LINE_RESET50,
                    "source_bits"   : 0,
                    "capture_bits"  : 0,
                    "identifier"    : "lr"}

LINE_RESET55 = """
repeat(30)              $no_clk_bit     0   0;
repeat(56)	            $clk_bit	    0	1; //Idle phase
        	            $clk_bit	    0	0; //Idle phase
"""
line_reset55 = {   "digipatsrc"    : LINE_RESET55,
                    "source_bits"   : 0,
                    "capture_bits"  : 0,
                    "identifier"    : "lr"}

#41 bits source (request, register value, (odd) parity), 0 bits capture
TEMPLATE_DATA_FROM_HOST = """
repeat(8), source       $clk_bit	    0	D; //host configuration byte
	                    $clk_bit	    0	X; //turnround
	                    $clk_bit	    0	H; //slave ack0
	                    $clk_bit	    0	L; //slave ack1
	                    $clk_bit	    0	L; //slave ack2
	                    $clk_bit	    0	X; //turnround
repeat(32), source	    $clk_bit	    0	D; //data[31:0]
source                  $clk_bit	    0	D; //parity bit from host
"""
template_data_from_host = {     "digipatsrc"    : TEMPLATE_DATA_FROM_HOST,
                                "source_bits"   : 41,
                                "capture_bits"  : 0,
                                "identifier"    : 'dfh'}

#8 bits source (request), 33 bits capture (register value, (odd) parity)
TEMPLATE_DATA_FROM_TARGET = """
repeat(8), source	    $clk_bit	    0	D; //host configuration byte
	                    $clk_bit	    0	X; //turnround
	                    $clk_bit	    0	H; //slave ack0
	                    $clk_bit	    0	L; //slave ack1
	                    $clk_bit	    0	L; //slave ack2
repeat(32), capture	    $clk_bit	    0	V; //data[31:0]
capture	                $clk_bit	    0	V; //parity bit from target
                        $clk_bit	    0	X; //turnround
"""
template_data_from_target = {   "digipatsrc"    : TEMPLATE_DATA_FROM_TARGET,
                                "source_bits"   : 8,
                                "capture_bits"  : 33,
                                "identifier"    : "dft" }


#trailer pattern needed by every pattern
TEMPLATE_TRAILER = """
jump_if(seqflag3, end)                      $wait_short -   -; //branch over dummy reads if none are needed
capture, repeat(reg14)         	            $no_clk_bit	-	V; //dummy capture to fill memory
end:
capture_stop	                            $no_clk_bit	-	-; //stop capturing
halt	                                    $no_clk_bit	-	-; //end of pattern
}
"""
template_trailer = {"digipatsrc"    : TEMPLATE_TRAILER,
                    "source_bits"   : 0,
                    "capture_bits"  : 0,
                    "identifier"    : 'trl'}


TEMPLATE_START_LOOP = """
set_loop(reg15)	        $wait_short     1   0; //set the loop
LoopStart:
"""
template_start_loop = { "digipatsrc"    : TEMPLATE_START_LOOP,
                        "source_bits"   : 0,
                        "capture_bits"  : 0,
                        "identifier"    : 'ls' }


TEMPLATE_END_LOOP = """
end_loop(LoopStart)     $wait_short     1   0; // loop end
"""
template_end_loop = {   "digipatsrc"    : TEMPLATE_END_LOOP,
                        "source_bits"   : 0,
                        "capture_bits"  : 0,
                        "identifier"    : 'le' }

def integerize( value ) -> int:
    """ automatically casts to integer
        If the value is a string and starts with '0x' or '0X', hex notation is assumed

    Args:
        value : the value to integerize

    Returns:
        int: the integer value
    """
    if type(value) == str:
        if "0X" in value.upper():
            value = int(value, 16)
        else:
            value = int(value)
    
    return value

class PXI_digital_pattern_SWD(ams_SWD_communication):
    """PXI_digital_pattern_SWD
    Implementation of the ams_SWD_communication interface on a PXIe digital pattern card
    """

    def _add_uint32_to_bitstream(self, bitstream: BitStream, a: int) -> BitStream :
        """_add_uint32_to_bitstream WORKAROUND method for a bug that prevents "BitStream" from appending a 32 bit unsigned integer

        Args:
            bitstream (BitStream): The bitstream where the uint32 shall be appended
            a (int): the uint32

        Returns:
            BitStream: the bitstream with the appended uint32
        """
        lower_word = a&0x0000FFFF
        upper_word = (a&0xFFFF0000) >> 16
        bitstream.write(uint16(upper_word), uint16)
        bitstream.write(uint16(lower_word), uint16)
        
        return bitstream

    def _add_nibble_to_bitstream(self, bitstream:BitStream, nib: int) -> BitStream:
        """add a nibble (4 bits) to a bitstream

        Args:
            bitstream (BitStream): The bitstream where the nibble shall be appended
            nib (int): the nibble

        Returns:
            BitStream: the bitstream with the appended nibble
        """
        # print(f"the nibble: {nib:x}")
        for i in range(4):
            bitstream.write((nib&(1<<(3-i)))>>(3-i), bool) #MSB first! 
        # print(f"the bitstream: {bitstream}")
        return bitstream

    def _read_uint32_parity_from_bitstream(self, bitstream:BitStream) -> Tuple[int, int]:
        """extract a 32-bit word and its corresponding parity bit from a bitstream

        Args:
            bitstream (BitStream): The bitstream from which the data should be read

        Returns:
            Tuple[int, int]: A tuple containing the 32 bit integer and the parity bit
        """
        the_uint32 = int(str(bitstream.read(32)),2)
        the_uint32 = self._reverse_bits(the_uint32)
        parity = int(str(bitstream.read(1)),2)
        return the_uint32, parity

    def _get_template_string(self, steps:list) -> str:
        """extracts the complete template pattern string for a given transaction

        Args:
            steps (list): A list of dictionaries containing the abstracted transaction steps

        Returns:
            str: the complete template pattern string
        """
        template = ""
        for step in steps:
            template += step["template"]["digipatsrc"]
        return template

    def _fill_pattern_template(self, template:str, namings: dict, dummy_captures:int) -> str: 
        """_fill_pattern_template Fills a pattern template with corresponding signal names, timing names, etc

        Args:
            template (str): the template pattern string to fill
            namings (dict): a dict of the namings to use in this pattern
            dummy_captures(int): the number of dummy captures in this pattern

        Returns:
            str: the .digipatsrc pattern filled with the needed signals
        """

        filled = template
        for timeset in self.timing_dict:
            swdLogger.debug(timeset)
            swdLogger.debug(self.timing_dict[timeset]["name"])
            filled = filled.replace("$"+timeset, self.timing_dict[timeset]["name"])
        filled = filled.replace("$pattern_name", namings["pattern_name"])
        filled = filled.replace("$swclk_channel", self.device_dict['Details']["settings"]['swclk'])
        filled = filled.replace("$swdio_channel", self.device_dict['Details']["settings"]['swdio'])
        filled = filled.replace("$source_wfm", namings["source_wfm"])
        filled = filled.replace("$capture_wfm", namings["capture_wfm"])
        # if (dummy_captures == 0):
        #     filled = filled.replace("$dummycapture_times", "")
        # else:
        #     filled = filled.replace("$dummycapture_times", f", repeat({dummy_captures})")
        return filled

    def _reverse_bits(self, a:int, length:int = 32) -> int:
        """_reverse_bits reverses a integer bitwise (MSB -> LSB)

        Args:
            a (int): The integer to reverse
            length (int): the bit length of the integer

        Returns:
            int: The reversed integer
        """
        result = 0
        for i in range(length):
            result <<= 1
            result |= a&1
            a >>= 1
        return result

    def _calc_parity(self, data: int, length: int = 32) -> int:
        """_calc_parity Calculates the the bit needed for odd parity

        Args:
            data (int): The integer data to calculate the parity on
            len (int): the length of the data to calculate the parity on. Defaults to 32

        Returns:
            int: The Parity bit
        """
        mask = 0x00000001
        cnt = 0
        temp_data = {}
        for i in range (length):
            temp_data[i] = (data&mask)
            if(temp_data[i] != 0):
                cnt = cnt + 1
            mask = mask << 1
        
        if (cnt % 2) == 0:
            return 0
        else: 
            return 1 

    def _calculate_packet_request(self, APnDP: int, RnW: int, address2_3: str) -> int:
        """calculates the packet request byte for given parameters

        Args:
            APnDP (int): select AP(1) or DP(0) request
            RnW (int): select Read(1) or Write(0) request
            address2_3 (str): AP or DP register address bits [2:3]

        Returns:
            int: the packet request byte for the given parameters
        """
        address2_3 = int(address2_3, 16)
        address_2 = (address2_3 & 0x02) >> 1
        address_3 = (address2_3 & 0x01) >> 0

        packet_request_byte = 0x80 #start bit set
        packet_request_byte |= (0x00, 0x40)[APnDP]    #AP/DP access
        packet_request_byte |= (0x00, 0x20)[RnW]      #read/write access
        packet_request_byte |= (address_2<<4)         #register address bit 2
        packet_request_byte |= (address_3<<3)         #register address bit 3
        packet_request_byte |= (0x00, 0x04)[self._calc_parity(packet_request_byte&0x7F, 8)]
        packet_request_byte |= 0x00   #stop bit, should be 0
        packet_request_byte |= 0x01   #park bit sent before changing SWDIO to open-drain, should be 1

        return packet_request_byte

    def _get_driver_id(self) -> str:
        """returns an id for the current configuration of the driver, based on  the channel names
        Returns:
            str: the id
        """
        return self.device_dict['Details']["settings"]['swclk'] + "_" + self.device_dict['Details']["settings"]['swdio']

    def _generate_pattern_id(self, pattern_src:dict):
        """generate a pattern id based on a list of swd transaction steps

        Args:
            pattern_src (dict): dictionary containing the list of abstracted steps
        """

        if (not "id" in pattern_src) or (not pattern_src["id"]):
            steps = pattern_src["steps"]
            id = ""
            for step in steps:
                id += step["template"]["identifier"]
                id += "_"

            pattern_src["id"] = id[:-1]

    def _calculate_capture_size(self, pattern_src:dict, no_loops:int=None):
        """ calculate the size of the capture memory needed and save it to the src dictionary

        Args:
            pattern_src (dict): dictionary containing the list of abstracted steps
            no_loops (int): number of loops

        Returns:
            Tuple[int, int]: the number of 32 bit words to read, the number of bits to capture to fill the last word
        """

        capture_bits = 0

        steps = pattern_src["steps"]

        for i in range(len( steps ) ):

            if steps[i]["template"]["identifier"] == "ls": #loop start
                # swdLogger.info(f"found loop start at index {i}")
                for j in range(no_loops - 1): #repeat for all loops except the last
                    k = i
                    while (steps[k]["template"]["identifier"] != "le"): #loop end
                        capture_bits += steps[k]["template"]["capture_bits"]
                        k += 1
                    # swdLogger.info(f"at loops end: {k}, iteration {j} ")
            else:
                capture_bits += steps[i]["template"]["capture_bits"]
            

        capture_words = math.ceil(float(capture_bits)/32.0)
        dummy_captures = capture_words *32 - capture_bits


        # swdLogger.info(f" CAPTURE WORDS {capture_words}")
        # swdLogger.info(f" CAPTURE BITS  {capture_bits}")
        # swdLogger.info(f" DUMMY CAPTURES {dummy_captures}")

        pattern_src["capture_words"] = capture_words
        pattern_src["dummy_captures"] = dummy_captures

    def _build_namings_timesets(self):
        """build the namings for all used timesets
        """
        for timeset in self.timing_dict:
            self.timing_dict[timeset]["name"] = "SWD_" + self._get_driver_id() + "_" + timeset

    def _build_naming(self, transaction: dict):
        """build the naming for a transaction

        Args:
            transaction (dict): the transaction abstraction
        """
        if not transaction["namings"]:
            pattern_name = "SWD_" + self._get_driver_id() + "_" + transaction["name"] + "_" + transaction["pattern_src"]["id"]
            transaction["namings"] = {
                                        #"transaction_name": transaction_name,
                                        "pattern_name": pattern_name,
                                        "source_wfm":   pattern_name + "_source",
                                        "capture_wfm":  pattern_name + "_capture",
                                        "pattern_path": os.path.join(self.device_dict['Details']['patterns_directory'], pattern_name + ".digipat")
                                    }

    def _create_timesets(self):
        """_create_timesets create all timesets needed by the swd driver
        """
        for timeset in self.timing_dict:
            self.device_dict['Details']['Session'].create_time_set(self.timing_dict[timeset]["name"])      

    def _configure_timesets(self):
        """_configure_timesets Set all edges of the timeset depending on the before-set configuration
        """
        period = 1/self.device_dict['Details']['speed']

        for timeset in self.timing_dict:
            if "period" in self.timing_dict[timeset]:
                self.device_dict['Details']['Session'].configure_time_set_period(self.timing_dict[timeset]["name"], self.timing_dict[timeset]["period"] * period)
                for channel in self.timing_dict[timeset]["edges"]:
                    self.device_dict['Details']['Session'].pins[self.device_dict['Details']["settings"][channel]]\
                        .configure_time_set_drive_edges(self.timing_dict[timeset]["name"],\
                                                        self.timing_dict[timeset]["edges"][channel]["format"],\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_on_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_data_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_return_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_off_edge"] * period)
                    self.device_dict['Details']['Session'].pins[self.device_dict['Details']["settings"][channel]]\
                        .configure_time_set_compare_edges_strobe(self.timing_dict[timeset]["name"],\
                                                                self.timing_dict[timeset]["edges"][channel]["strobe_edge"] * period)
            elif "duration" in self.timing_dict[timeset]:
                period = self.timing_dict[timeset]["duration"]
                self.device_dict['Details']['Session'].configure_time_set_period(self.timing_dict[timeset]["name"],  period)
                for channel in self.timing_dict[timeset]["edges"]:
                    self.device_dict['Details']['Session'].pins[self.device_dict['Details']["settings"][channel]]\
                        .configure_time_set_drive_edges(self.timing_dict[timeset]["name"],\
                                                        self.timing_dict[timeset]["edges"][channel]["format"],\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_on_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_data_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_return_edge"] * period,\
                                                        self.timing_dict[timeset]["edges"][channel]["drive_off_edge"] * period)
                    self.device_dict['Details']['Session'].pins[self.device_dict['Details']["settings"][channel]]\
                        .configure_time_set_compare_edges_strobe(self.timing_dict[timeset]["name"],\
                                                                self.timing_dict[timeset]["edges"][channel]["strobe_edge"] * period)
            else:
                swdLogger.error(f"timeset not properly defined: {timeset}")
            
    def _set_voltage_levels(self):
        """_set_voltage_levels Set The voltage levels depending on the before-set configuration
        """
        v_il_ratio = 0
        v_ih_ratio = 1
        v_ol_ratio = 0.3
        v_oh_ratio = 0.7
        v_term_ratio = 1
        self.device_dict['Details']['Session']\
            .pins[self.device_dict['Details']["settings"]['swdio'], self.device_dict['Details']["settings"]['swclk']]\
            .configure_voltage_levels(  v_il_ratio * self.device_dict['Details']["bus_voltage"],\
                                        v_ih_ratio * self.device_dict['Details']["bus_voltage"],\
                                        v_ol_ratio * self.device_dict['Details']["bus_voltage"],\
                                        v_oh_ratio * self.device_dict['Details']["bus_voltage"],\
                                        v_term_ratio * self.device_dict['Details']["bus_voltage"]  )
        pass

    def _create_waveforms(self, transaction: dict):
        """create the needed waveforms for a transaction

        Args:
            transaction (dict): the transaction that the waveforms are generated for
        """
        if not transaction["namings"]["capture_wfm"] in self.capture_waveforms:
            self.device_dict['Details']['Session'].channels[self.device_dict['Details']["settings"]['swdio']].create_capture_waveform_serial(transaction["namings"]["capture_wfm"], 32, nidigital.BitOrder.MSB)
            self.capture_waveforms.append(transaction["namings"]["capture_wfm"])
        if not transaction["namings"]["source_wfm"] in self.source_waveforms:
            self.device_dict['Details']['Session'].channels[self.device_dict['Details']["settings"]['swdio']].create_source_waveform_serial(transaction["namings"]["source_wfm"], nidigital.SourceDataMapping.BROADCAST, 32, nidigital.BitOrder.MSB)
            self.source_waveforms.append(transaction["namings"]["source_wfm"])
        pass

    def _build_transaction(self, transaction:dict, force_recompile:bool = False, no_loops:int = None):
        """process and compile the pattern and set up waveforms for the given transaction

        Args:
            transaction (dict): The transaction abstraction
            force_recompile (bool, optional): On True:Compile the patterns even if they exist on disk. Defaults to False.
            no_loops (int): number of loops
        """
        #preprocess pattern
        self._generate_pattern_id(transaction["pattern_src"])
        self._calculate_capture_size(transaction["pattern_src"], no_loops)
        #build namings if needed
        self._build_naming(transaction)

        #create the waveforms
        self._create_waveforms(transaction)
        
        #make the patterns available
        if (not os.path.exists(transaction["namings"]["pattern_path"]) ) or force_recompile:
            self._compile_pattern(transaction)

    def _compile_pattern(self, transaction:dict):
        """ generate and compile pattern

        Args:
            transaction (dict): The transaction abstraction
        """

        template = self._get_template_string(transaction["pattern_src"]["steps"])      
        pattern = self._fill_pattern_template(template, transaction["namings"], transaction["pattern_src"]["dummy_captures"])
        
        # write the pattern to a file
        with open(transaction["namings"]["pattern_path"] + "src", "w") as file:
            file.write(pattern)
        # compile the pattern
        ams_ni_file_tools.compileDigipatSource(transaction["namings"]["pattern_path"] + "src", self.device_dict['Details']["ni_pin_map_file"])

    def _load_pattern(self, pattern_path:str):
        """load a pattern from a given path into the digital pattern instrument

        Args:
            pattern_path (str): path to the compiled pattern file
        """
        if not pattern_path in self.loaded_patterns:
            self.device_dict['Details']['Session'].load_pattern(pattern_path)
            self.loaded_patterns.append(pattern_path)

    def _fill_source_memory(self, transaction:dict, looped_data:List[int] = None, no_loops:int = None):
        """Load data defined in the transaction abstraction into the memory of the digital pattern instrument

        Args:
            transaction (dict): the transaction abstraction
        """
        if looped_data:
            data_iterator = iter(looped_data)
        the_bitstream = BitStream()
        #for step in transaction["pattern_src"]["steps"]:
        
        def fill_step(step):
            if "switch_code" in step:       #switch to swd
                the_bitstream.write(step["switch_code"], uint16)
            elif "packet_request" in step:  #data from/to target
                # print(f"bitstream before: {the_bitstream}")
                # print(f"packet request: {step['packet_request']:x}")
                the_bitstream.write(step["packet_request"], uint8)
                # print(f"bitstream after: {the_bitstream}")
                if "write_data" in step:
                    if type(step["write_data"]) == int:    
                        self._add_uint32_to_bitstream(the_bitstream, self._reverse_bits(step["write_data"]))
                        the_bitstream.write(self._calc_parity(step["write_data"]), bool)
                    elif step["write_data"] == "loop_data":
                        data = next(data_iterator)
                        self._add_uint32_to_bitstream(the_bitstream, self._reverse_bits(data))
                        the_bitstream.write(self._calc_parity(data), bool)

            elif "nibble" in step:          #generic output nibble
                self._add_nibble_to_bitstream(the_bitstream, step["nibble"])
            elif "byte" in step:          #generic output nibble
                the_bitstream.write(step["byte"], uint8)

        steps = transaction["pattern_src"]["steps"]
        for i in range(len( steps ) ):

            if steps[i]["template"]["identifier"] == "ls": #loop start
                #swdLogger.info(f"found loop start at index {i}")
                for j in range(no_loops - 1): #repeat for all loops except the last
                    k = i
                    while (steps[k]["template"]["identifier"] != "le"): #loop end
                        fill_step(steps[k])
                        k += 1
                    # swdLogger.info(f"at loops end: {k}, iteration {j} ")
            else:
                fill_step(steps[i])

        self._write_bitstream_to_source_memory(the_bitstream, transaction["namings"]["source_wfm"])

    def _write_bitstream_to_source_memory(self, the_bitstream:BitStream, the_waveform:str):
        """write a bitstream to the source memory of the ppmu instrument

        Args:
            the_bitstream (BitStream): The bitstream that should be written
            the_waveform (str): the name of the waveform the bitstream shall be written to
        """
        #build uint32 array representing the bitstream
        list_of_ints = []
        while True:
            bitsLeft = len(the_bitstream)
            if(bitsLeft > 32):
                newInt = the_bitstream.read(32)
                list_of_ints.append(int(str(newInt), 2)) 
            else:
                newInt = the_bitstream.read(bitsLeft)
                list_of_ints.append(int(str(newInt), 2)<<(32-bitsLeft))
                break

        #load source waveform into the instrument
        # print(the_waveform)
        self.device_dict['Details']['Session'].write_source_waveform_broadcast(the_waveform, list_of_ints)
        
    def _read_capture_memory(self, transaction:dict, no_loops:int = None) -> dict:
        """Read out the capture memory of the digital pattern instrument and extract the desired fields

        Args:
            transaction (dict): The transaction abstraction

        Returns:
            dict: a dict containing the read fields as defined in the abstraction, each split into a dict containing data and parity
        """
        return_dict = {}

        #fetch the capture
        captured_waveform_memory_view_per_site = self.device_dict['Details']['Session'].fetch_capture_waveform(transaction["namings"]["capture_wfm"], transaction["pattern_src"]["capture_words"])
        captured_waveform_words = captured_waveform_memory_view_per_site[0]

        # swdLogger.debug("captured: "+captured_waveform_words.hex())

        #rearrange captured data into right format
        the_bitstream = BitStream()
        for word in captured_waveform_words:
            the_bitstream.write(word, uint32)

        loop_read_increment = 0

        # swdLogger.info(f"number of loops {no_loops}")

        def read_step(step):
            nonlocal return_dict
            nonlocal loop_read_increment
            if "read_data_name" in step:
                read_data_name = step["read_data_name"]
                if (read_data_name == "loop_data"):
                    read_data_name = f"loop_data_{loop_read_increment}"
                    loop_read_increment += 1
                data, parity = self._read_uint32_parity_from_bitstream(the_bitstream)
                return_dict[read_data_name] = {}
                return_dict[read_data_name]["data"] = data
                return_dict[read_data_name]["parity"] = parity

        steps = transaction["pattern_src"]["steps"]
        for i in range(len( steps ) ):

            if steps[i]["template"]["identifier"] == "ls": #loop start
                #swdLogger.info(f"found loop start at index {i}")
                for j in range(no_loops - 1): #repeat for all loops except the last
                    k = i
                    while (steps[k]["template"]["identifier"] != "le"): #loop end
                        read_step(steps[k])
                        k += 1
                    # swdLogger.info(f"at loops end: {k}, iteration {j} ")
            else:
                read_step(steps[i])

        return return_dict

    def _check_parities(self, read_dict: dict) -> bool:
        """Bulk validate parities read after a read

        Args:
            read_dict (dict): the captured data, as returned by _read_capture_memory()

        Returns:
            bool: True if all parities are okay, False if one or more parity checks failed
        """
        ok = True
        for value in read_dict:
            ok = ok and ( read_dict[value]["parity"] == self._calc_parity(read_dict[value]["data"]) )
        return ok

    def _run_transaction(self, transaction:dict, looped_data_write:List[int] = [], no_loops: int = None, force_recompile = False) -> Tuple[bool, dict]:
        """runs a transaction

        Args:
            transaction (dict): the transaction, abstracted in its dict
            looped_data (List[int]): the data for a transaction containing a loop
            no_loops (int): number of loops

        Returns:
            Tuple[bool, dict]: first element: true on pattern success, second element: a dict containing all read values
        """

        swdLogger.debug("running transaction " + transaction["name"] + " ....")

        #check if pattern exists, build and compile otherwise
        self._build_transaction(transaction, force_recompile=force_recompile, no_loops=no_loops)

        self._load_pattern(transaction["namings"]["pattern_path"])

        #fill memory
        self._fill_source_memory(transaction, looped_data_write, no_loops)

        #tell the sequencer wether to read dummies or not (seqflag3)
        if (transaction["pattern_src"]["dummy_captures"] == 0):
            self.device_dict['Details']["Session"].write_sequencer_flag( nidigital.SequencerFlag.FLAG3, True) #True = read no dummies
        else:
            self.device_dict['Details']["Session"].write_sequencer_flag( nidigital.SequencerFlag.FLAG3, False) #True = read no dummies
        
        #set dummy capture register (reg14)
        self.device_dict['Details']["Session"].write_sequencer_register( nidigital.SequencerRegister.REGISTER14, transaction["pattern_src"]["dummy_captures"] )

        #burst
        pattern_passfail = self.device_dict['Details']['Session'].burst_pattern(transaction["namings"]["pattern_name"], timeout=30)[0]

        #read memory
        read_dict = self._read_capture_memory(transaction, no_loops)
        parity_passfail = self._check_parities(read_dict)

        swdLogger.debug("pattern pass/fail: " + str(pattern_passfail))
        swdLogger.debug("parity pass/fail: " + str(parity_passfail))

        passfail = pattern_passfail and parity_passfail

        #return
        return passfail, read_dict

    def _calculate_select_register(self, address:int, bank:int) -> Tuple[int, str]:
        """calculates the select register+additional address bytes and formats them for AP requests

        Args:
            address (int): The address of the AP
            bank (int): the bank to read/write from

        Returns:
            Tuple[int, str]: select register, address[2:3]
        """
        #build select register
        select_reg = 0
        #ap address
        select_reg |= (address<<24)
        #bank
        select_reg |= (bank&0xF0)
        
        #calculate address2_3 string
        address2_3 = hex(((address&0x4)>>2) | ((address&0x08) >> 3))

        return(select_reg, address2_3)

    def _read_ap(self, address:int, bank:int):      
        select_reg, address2_3 = self._calculate_select_register(address, bank)
        transaction = {   
            "name": "read_ap",
            "namings": None,
            "pattern_src":{
                "steps":    [   
                    {   "template": template_header   },
                    {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(0, 0, "0x1"), "write_data": select_reg },         #write SELECT
                    {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(1, 1, address2_3), "read_data_name":"dummy"},   #forward to AP, commit transaction
                    {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x11"), "read_data_name":"rdbuff"},      #read rdbuff   
                    {   "template": template_trailer }]}}

        the_return = self._run_transaction(transaction)
        if (the_return[0]):
            rdbuff = the_return[1]["rdbuff"]["data"]
            return f"0x{rdbuff:08x}"
        else:
            return "error"

    def _write_ap(self, address:int, bank:int, value:int):
        select_reg, address2_3 = self._calculate_select_register(address, bank)
        transaction = {   
            "name": "write_ap",
            "namings": None,
            "pattern_src":{
                "steps":    [   
                    {   "template": template_header   },
                    {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(0, 0, "0x1"), "write_data": select_reg },         #write SELECT
                    {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, address2_3), "write_data": value},        #forward to AP, write value
                    {   "template": template_trailer }]}}
        self._run_transaction(transaction)
            
    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: SWD object
        """
        self.device_dict = device_dict
        if not 'speed' in device_dict['Details']:
            swdLogger.warning("no SWD speed provided, using default: 60kHz")
            self.device_dict['Details']['speed'] = 60000

        if not 'patterns_directory' in device_dict['Details']:
            swdLogger.warning(r"no directory for created digital patterns provided, using default: 'C:\NIDIGITALPATTERNS'")
            self.device_dict['Details']['patterns_directory'] = r"C:\NIDIGITALPATTERNS"
        #make sure the patterns directory exists
        if not os.path.exists(self.device_dict['Details']['patterns_directory']):
            os.makedirs(self.device_dict['Details']['patterns_directory'])    

        #drive edges and periods are given as ratios compared to the period defined by the device dict speed
        # 1 = full period
        self.timing_dict = {
            "wait":{        "name":None, 
                            "period": 0.2,  
                            "edges":{       "swdio":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "swclk":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 1,
                                                        "drive_data_edge": 1,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},

            "wait_short":{  "name":None, 
                            "duration": 1e-06,  
                            "edges":{       "swdio":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "swclk":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}},

            "clk_bit":{     "name":None,
                            "period": 1,
                            "edges":{       "swdio":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 0.98,
                                                        "strobe_edge": 0.5     },
                                            "swclk":{   "format": nidigital.DriveFormat.RH,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0.5,
                                                        "drive_return_edge": 0.98,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.25     }}},

            "no_clk_bit":{  "name":None,
                            "period": 1,
                            "edges":{       "swdio":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     },
                                            "swclk":{   "format": nidigital.DriveFormat.NR,
                                                        "drive_on_edge": 0,
                                                        "drive_data_edge": 0,
                                                        "drive_return_edge": 1,
                                                        "drive_off_edge": 1,
                                                        "strobe_edge": 0.5     }}}
        }

        self.transactions_dict = {
            "init": {   "name": "init",
                        "namings":None,
                        "pattern_src":{
                            "steps":    [   
                                {   "template": template_header   },
                                {   "template": template_switch_to_swd, "switch_code": 0x79E7},
                                {   "template": template_idle },
                                {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x0"), "read_data_name":"idcode"},          #read IDCODE
                                {   "template": template_idle },
                                {   "template": template_idle },
                                {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(0, 0, "0x1"), "write_data": 0x000000F0 },            #write SELECT to AHB-AP, bank 0F
                                {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(0, 0, "0x2"), "write_data": 0x50000000 },            #write PWRUP request to CTRL/STAT
                                {   "template": template_idle },
                                {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x2"), "read_data_name":"ctrlstat_1"},      #read CTRL/STAT
                                {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(1, 1, "0x3"), "read_data_name":"dummy"},           #dummy read IDR AP0
                                {   "template": template_idle },
                                {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(1, 1, "0x3"), "read_data_name":"idr"},             #read IDR AP0 for real
                                {   "template": template_idle },
                                {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(0, 0, "0x1"), "write_data": 0x00000000 },            #write SELECT to AHB-AP0
                                {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(1, 0, "0x0"), "write_data": 0x23000052 },            #enable MEM-AP and 32 bit access(write CSW)
                                {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x2"), "read_data_name":"ctrlstat_2"},      #read CTRL/STAT
                                {   "template": template_trailer },]}},
            "init_aeon": {   "name": "init_aeon",
                        "namings":None,
                        "pattern_src":{
                            "steps":    [   
                                {   "template": template_header   },
                                {   "template": line_reset55},
                                {"template": template_output_byte, "byte": 0xFA},
                                {"template": template_output_byte, "byte": 0xEE},
                                {"template": template_output_byte, "byte": 0xEE},
                                {"template": template_output_byte, "byte": 0xEE},
                                {"template": template_output_byte, "byte": 0x6F},
                                {"template": template_output_byte, "byte": 0xF4},
                                {"template": template_output_byte, "byte": 0x9C},
                                {"template": template_output_byte, "byte": 0xF9},
                                {"template": template_output_byte, "byte": 0x04},
                                {"template": template_output_byte, "byte": 0x6A},
                                {"template": template_output_byte, "byte": 0x9B},
                                {"template": template_output_byte, "byte": 0x4A},
                                {"template": template_output_byte, "byte": 0x16},
                                {"template": template_output_byte, "byte": 0x19},
                                {"template": template_output_byte, "byte": 0x7F},
                                {"template": template_output_byte, "byte": 0x5B},
                                {"template": template_output_byte, "byte": 0xBC},
                                {"template": template_output_byte, "byte": 0x74},
                                {"template": template_output_byte, "byte": 0x57},
                                {"template": template_output_byte, "byte": 0x03},
                                {"template": template_output_byte, "byte": 0xD9},
                                {"template": template_output_byte, "byte": 0x80},
                                {"template": template_output_byte, "byte": 0x58},
                                {   "template": line_reset50},
                                {   "template": template_idle },
                                {   "template": template_data_from_target, "packet_request": 0xA5, "read_data_name":"idcode"},          #read IDCODE
                                {   "template": template_data_from_host, "packet_request": 0x95, "write_data": 0x50000000},          #addr. 0x95 - PWRUP request                                
                                {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"CTRL/STAT"},          #addr. 0xB1 - READ CTRL/STAT
                                {   "template": template_data_from_host, "packet_request": 0x95, "write_data": 0x50000000},          #addr. 0x95 - PWRUP request
                                {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"CTRL/STAT_2"},          #addr. 0xB1 - READ CTRL/STAT
                                {   "template": template_data_from_host, "packet_request": 0x8D, "write_data": 0x00000000},          # addr. x8D - write SELECT to AHB-AP0
                                {   "template": template_idle },    # 6 us delay
                                {   "template": template_data_from_host, "packet_request": 0xDD, "write_data": 0x6B587032},         # Write AP DRW with key
                                {   "template": template_data_from_host, "packet_request": 0xC5, "write_data": 0x4528482B},         # Write AP DRW with key                                
                                {   "template": template_data_from_host, "packet_request": 0xC9, "write_data": 0x73357638},         # Write AP DRW with key                              
                                {   "template": template_data_from_host, "packet_request": 0xD1, "write_data": 0x792F423F},         # Write AP DRW with key      
                                {   "template": template_data_from_host, "packet_request": 0xD1, "write_data": 0x792F423F},         # Write AP DRW with key
                                {   "template": template_trailer },]}},
            "read_register":    {   "name": "read_register",
                                    "namings": None,
                                    "pattern_src":{
                                        "steps":    [   
                                            {   "template": template_header   },
                                            {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(1, 0, "0x2"), "write_data": None },                  #write TAR
                                            {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x2"), "read_data_name":"ctrlstat"},        #read CTRL/STAT
                                            {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(1, 1, "0x3"), "read_data_name":"dummy"},           #dummy read from drw
                                            {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(1, 1, "0x3"), "read_data_name":"drw"},             #actual read from drw
                                            {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x3"), "read_data_name":"rdbuff"},          #read from RDBUFF
                                            {   "template": template_trailer },]}},
            "write_register":   {   "name": "write_register",
                                    "namings": None,
                                    "pattern_src":{
                                        "steps":    [   
                                            {   "template": template_header   },
                                            {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(1, 0, "0x2"), "write_data": None },                  #write TAR
                                            {   "template": template_data_from_host, "packet_request": self._calculate_packet_request(1, 0, "0x3"), "write_data": None },                  #write DRW
                                            {   "template": template_data_from_target, "packet_request": self._calculate_packet_request(0, 1, "0x2"), "read_data_name":"ctrlstat"},        #read CTRL/STAT
                                            {   "template": template_trailer },]}},
            "write_frame":  {   "name": "write_frame",
                                "namings": None,
                                "pattern_src":{
                                    "steps":    [   
                                        {   "template": template_header   },
                                        {   "template": template_data_from_host, "packet_request": None, "write_data": None },
                                        {   "template": template_trailer },]}},
            "read_frame":   {   "name": "read_frame",
                                "namings": None,
                                "pattern_src":{
                                    "steps":    [   
                                        {   "template": template_header   },
                                        {   "template": template_data_from_target, "packet_request": None, "read_data_name":"read_data"},
                                        {   "template": template_trailer },]}}
        }
        self.source_waveforms = []
        self.capture_waveforms = []
        self.loaded_patterns = []

    def SWD_init(self) -> bool:
        """
        This method initializes the SWD driver on a digital pattern instrument

        Returns:
            Bool: True if the Initialization worked, False otherwise
        """       
        #add timing for swd communications
        try:
            self._build_namings_timesets()
            self._create_timesets()
            self._configure_timesets()

            #set the voltage levels
        except nidigital.Error:
            pass
        self._set_voltage_levels()

        self.source_waveforms = []
        self.capture_waveforms = []
        self.loaded_patterns = []
        return True

    def SWD_configure(
        self, 
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
        ):
        """SWD_configure 
        This method configures the SWD communication.
        If this method is not called the instance uses the defaults given in the HW config file

        Args:
            speed (int): speed of the communication
            bus_voltage (float): bus voltage level in V
            settings_dict (dict): defines any setting that may be implementation specific as interface pins in case of PXI pattern instrument. Defaults to {}.
        """

        if not self.device_dict['Details']['speed'] == speed:
            self.device_dict['Details']['speed'] = speed
            self.__configure_timesets()

        if not self.device_dict['Details']["bus_voltage"] == bus_voltage:
            self.device_dict['Details']["bus_voltage"] = bus_voltage
            self.__set_voltage_levels()

        if settings_dict:
            recompile = False
            if 'swdio' in settings_dict:
                if not (self.device_dict['Details']["settings"]['swdio'] == settings_dict['swdio']):
                    self.device_dict['Details']["settings"]['swdio'] = settings_dict['swdio']
                    recompile = True

            if 'swclk' in settings_dict:
                if not (self.device_dict['Details']["settings"]['swclk'] == settings_dict['swclk']):    
                    self.device_dict['Details']["settings"]['swclk'] = settings_dict['swclk']
                    recompile = True      

            if recompile:
                self._build_namings()
                self._build_patterns()
                self._create_waveforms()
                self._load_patterns()
        pass

    def SWD_pre_init_pattern(self, data: list):
        """SWD_pre_init_pattern 
        This method gives access to the bus and any initial pattern can be sent

        Args:
            data (list): string of characters describing a bitstream with hex values like: 5575AC127E
                there is no maximum length defined for this argument
        """

        transaction = {     "name": "pre_init_pattern",
                            "namings": None,
                            "pattern_src":{
                                "steps":       
                                      [ {   "template": template_header   } ]
                                    + [ {   "template": template_output_nibble, "nibble":int(a, 16)} for a in data]
                                    + [ {   "template": template_trailer }]}}

        self._run_transaction(transaction)

    def SWD_init_sequence(self, d_idcode: int, d_idr: int) -> bool:
        """SWD_init_sequence This initializes the SWD target

        Args:
            d_idcode (int): The IDCODE that should be read from the target
            d_idr (int): The IDR that should be read from the target

        Returns:
            bool: True on success, False otherwise
        """
        swdLogger.info("initializing swd interface on pins: (SWDIO:"+self.device_dict['Details']["settings"]['swdio']+",SWCLK:"+self.device_dict['Details']["settings"]['swclk']+")")

        pattern_ok, return_dict = self._run_transaction(self.transactions_dict["init"])

        values_ok = True
        values_ok = values_ok and (d_idcode == return_dict["idcode"]["data"])
        values_ok = values_ok and (d_idr == return_dict["idr"]["data"])

        return values_ok and pattern_ok
       
    def SWD_init_sequence_aeon(self) -> bool:
        """
        This method generates the initializing sequence with keyword unlock for Aeon on SWD.

        Args:
            d_idcode (int): ID code to check
            key_1 (int): first segment of the keyword
            key_2 (int): second segment of the keyword
            key_3 (int): third segment of the keyword
            key_4 (int): fourth segment of the keyword

        Returns:
            bool: it is true if all steps executed properly
        """
        swdLogger.info("initializing swd interface on pins: (SWDIO:"+self.device_dict['Details']["settings"]['swdio']+",SWCLK:"+self.device_dict['Details']["settings"]['swclk']+")")

        pattern_ok, return_dict = self._run_transaction(self.transactions_dict["init_aeon"])
        # print(pattern_ok)


        # values_ok = True
        
        swdLogger.debug(f'IDCODE: {return_dict["idcode"]["data"]:x}')
        swdLogger.debug(f'CTRL/STAT :{return_dict["CTRL/STAT"]["data"]:x}')
        swdLogger.debug(f'CTRL/STAT 2: {return_dict["CTRL/STAT_2"]["data"]:x}')
        swdLogger.debug(f"Pattern OK (all ACK positive): {pattern_ok}")

        return pattern_ok
        
    def SWD_send_frame(
        self, 
        APnDP: int, 
        RnW: int, 
        address2_3: str, 
        payload: str=''
        ) -> str:
        """SWD_send_frame 
        This method builds SWD frame based on interface spec by standard parameters

        Args:
            APnDP (int): A single bit, indicating whether the Debug Port or the Access Port Access Register is to be accessed. 1 for accessing AP
            RnW (int): A single bit, this bit is 0 for an write access, or 1 for a read access.
            address2_3 (str): Two bits, giving the A[3:2] address field for the DP or AP register address.
            payload (str): payload of the request - if it is not given then reading is going to be done

        Returns:
            str: returns the bytes received from the bus in case of reading
        """

        if (not RnW and len(payload) == 0):
            swdLogger.error("SWD_send_frame - RnW set to Write, but no payload provided")
            pass

        packet_request_byte = self._calculate_packet_request(APnDP, RnW, address2_3)
        # print(f'{packet_request_byte:x}'.upper())
        if (RnW or len(payload) == 0):
            #read frame
            #fill data into packet abstraction
            self.transactions_dict["read_frame"]["pattern_src"]["steps"][1]["packet_request"] = packet_request_byte    
            #burst pattern
            transaction_passfail, transaction_return = self._run_transaction(self.transactions_dict["read_frame"])
            return hex(transaction_return["read_data"]["data"])
        else:
            #write frame
            #convert data string into integer
            payload_int = int(payload, 16)
            #fill data into packet abstraction
            self.transactions_dict["write_frame"]["pattern_src"]["steps"][1]["packet_request"] = packet_request_byte
            self.transactions_dict["write_frame"]["pattern_src"]["steps"][1]["write_data"] = payload_int      
            #burst pattern
            transaction_passfail, transaction_return = self._run_transaction(self.transactions_dict["write_frame"])
            pass

        pass

    def SWD_send_DP_request(
        self, 
        req_type: ams_SWD_communication.ams_SWD_DP_req_type=ams_SWD_communication.ams_SWD_DP_req_type.R_IDCODE, 
        payload: str=''
        ) -> str:
        """SWD_send_DP_request 
        This method sends DP request
        If payload is not defined at the call, read requiest is going to be sent

        Args:
            req_type (ams_SWD_DP_req_type, optional): type of the request. Defaults to ams_SWD_DP_req_type.R_IDCODE.
            payload (str, optional): payload of the request, in case of reading this field should not be given. Defaults to ''.

        Returns:
            str: returns the bytes received from the bus in case of read request
        """
        register_address = int(req_type.value, 16)
        address2_3 = hex((register_address & 0x0C)>>2)
        RnW = (0,1)[(payload=='')]

        return self.SWD_send_frame(0, RnW, address2_3, payload)

    def SWD_send_AP_request(
        self, 
        ap_address: str,
        word_bank: str,
        payload: str=''
        ) -> str:
        """SWD_send_AP_request 
        This method sends DP request
        If payload is not defined at the call, read requiest is going to be sent

        Args:
            req_type (ams_SWD_AP_req_type, optional): type of the request. Defaults to ams_SWD_AP_req_type.IDR.
            payload (str, optional): payload of the request, in case of reading this field should not be given. Defaults to ''.

        Returns:
            str: returns the bytes received from the bus in case of read request
        """
        address = int(ap_address, 16)
        bank = int(word_bank, 16)
        if payload != '':
            value = int(payload,16)
            self._write_ap(address, bank, value)
            return ''
        else:
            return self._read_ap(address, bank)        

    def SWD_write_register(
        self, 
        address: str, 
        data: str
        ) -> bool:
        """SWD_write_register 
        Implements the register writing procedure via SWD interface

        Args:
            address (str): memory address to be written
            data (str): memory content to be sent

        Returns:
            bool: returns true if the task is done without any problem
        """

        #convert data/address strings into integers
        address_int = int(address, 16)
        data_int = int(data, 16)

        #fill data into packet abstraction
        self.transactions_dict["write_register"]["pattern_src"]["steps"][1]["write_data"] = address_int
        self.transactions_dict["write_register"]["pattern_src"]["steps"][2]["write_data"] = data_int
        
        #burst pattern
        transaction_passfail, transaction_return = self._run_transaction(self.transactions_dict["write_register"])

        #check the values
        status_good = True
        if (transaction_return["ctrlstat"]["data"] != 0xF0000040):
            swdLogger.error("ctrl/stat in swd read not valid!; must be: " + hex(0xF0000040) + "; actual: "+ hex(transaction_return["ctrlstat"]["data"]))
            status_good = False
        if (status_good):
            swdLogger.debug("ctrl/stat in swd read good")

        success = transaction_passfail and status_good

        return success

    def SWD_read_register(
        self, 
        address: str
        ) -> str:
        """SWD_read_register 
        Implements the register reading procedure via SWD interface

        Args:
            address (str): memory address to be read

        Returns:
            str: content of the addressed memory
        """

        #convert str address to int
        address_int = int(address, 16)

        #fill data into packet abstraction
        self.transactions_dict["read_register"]["pattern_src"]["steps"][1]["write_data"] = address_int

        #burst pattern
        read_passfail, transaction_return = self._run_transaction(self.transactions_dict["read_register"])

        swdLogger.debug("read DRW: " + hex(transaction_return["drw"]["data"]))
        swdLogger.debug("read RDBUFF: " + hex(transaction_return["rdbuff"]["data"]))

        #check the values
        status_good = True
        if (transaction_return["ctrlstat"]["data"] != 0xF0000040):
            swdLogger.error("ctrl/stat in swd read not valid!; must be: " + hex(0xF0000040) + "; actual: "+ hex(transaction_return["ctrlstat"]["data"]))
            status_good = False
        if (status_good):
            swdLogger.debug("ctrl/stat in swd read good")

        success = read_passfail and status_good
        #return true only if everything is okay, also return the read value
        #return (success, transaction_return["drw"]["data"])

        #format data as string for return
        return hex(transaction_return["drw"]["data"])

    def SWD_parity(
        self, 
        data: str
        ) -> bool:
        """
        This method is responsible to calculate the parity bit

        Return:
            bool: parity bit value
        """
        data_int = int(data, 16)
        return self._calc_parity(data_int)

    def SWD_bulk_read(
        self,
        bank:int,
        csw_format:int,
        addresses:List) -> Tuple[int, List[int]]:

        bank = integerize(bank)
        csw_format = integerize(csw_format)

        addresses = [integerize(address) for address in addresses]

        swdLogger.debug(f"bulk reading {len(addresses)} registers")

        transaction = {     "name": f"bulk_read_{len(addresses)}",
                            "namings": None,
                            "pattern_src":{
                                "steps":       
                                      [ {   "template": template_header   },
                                        {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E  }, #send abort request
                                        {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank  }, #select bank
                                        {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":csw_format}, #set csw format
                                        {   "template": template_start_loop }, #start the loop
                                        {   "template": template_data_from_host, "packet_request": 0xD1, "write_data":"loop_data" }, #set address
                                        {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"loop_data"}, #read back drw
                                        {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"loop_data"}, #read back drw
                                        {   "template": template_data_from_target, "packet_request": 0xBD, "read_data_name":"loop_data"}, #read back rdbuff -> contains the value of interest
                                        {   "template": template_end_loop}, #end the loop
                                        {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"loop_data"}, #read back ctrl_stat
                                        {   "template": template_trailer }]}}

        no_loops = len(addresses)

        #setup loop count for vector execution. the loops in this driver use reg15 for loop count
        self.device_dict['Details']["Session"].write_sequencer_register( nidigital.SequencerRegister.REGISTER15, no_loops )

        passfail, transaction_return = self._run_transaction(transaction, addresses, no_loops, force_recompile=False)

        read_memory = [transaction_return[f"loop_data_{i*3 + 2}"]["data"] for i in range(len(addresses))]

        return passfail, read_memory

    def SWD_bulk_write(
        self,
        bank:int,
        csw_format:int,
        addresses:List[int],
        data:List[int]) -> bool:

        bank = integerize(bank)
        csw_format = integerize(csw_format)
        addresses = [integerize(address) for address in addresses]
        data = [integerize(datum) for datum in data]

        #check basic validness
        if len(addresses) != len(data):
            swdLogger.error("not the same number of addresses and corresponding data provided!")
            return False

        transaction = {     "name": "bulk_write",
                            "namings": None,
                            "pattern_src":{
                                "steps":       
                                      [ {   "template": template_header   },
                                        {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat_pre_write"},
                                        {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E}, #abort request
                                        {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank},       #banksel
                                        {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":0x23000012}, #CSW
                                        {   "template": template_start_loop }, #start the loop
                                        {   "template": template_data_from_host, "packet_request": 0xD1, "write_data": "loop_data"}, #set TAR
                                        {   "template": template_data_from_host, "packet_request": 0xDD, "write_data": "loop_data"}, #set the data
                                        {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name": "loop_data"}, #dummy read to initiate transfer
                                        {   "template": template_end_loop}, #end the loop
                                        {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat_post_write"},
                                        {   "template": template_trailer }] }}

        no_loops = len(addresses)
        #bring address and data into format
        # loop_data_write = np.zeros(1600000, dtype=int)
        loop_data_write = []
        for i in range(no_loops):
            loop_data_write.append(addresses[i])
            loop_data_write.append(data[i])
            # loop_data_write[i*2] = addresses[i]
            # loop_data_write[i*2 + 1] = data[i]

        #setup loop count for vector execution. the loops in this driver use reg15 for loop count
        self.device_dict['Details']["Session"].write_sequencer_register( nidigital.SequencerRegister.REGISTER15, no_loops )

        passfail, transaction_return = self._run_transaction(transaction, list(loop_data_write), no_loops)

        return passfail


    def reinit(self):
        self.__init__(self.device_dict)


'''
    # def SWD_read_core_status(
    #     self
    #     ) -> Tuple[bool, Dict]:
    #     swdLogger.debug("reading core status")
    #     transaction = {     "name": "read_ctrl_stat",
    #                         "namings": None,
    #                         "pattern_src":{
    #                             "steps":       
    #                                   [ {   "template": template_header   },
    #                                     {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat"},
    #                                     {   "template": template_trailer} ] }}

    #     passfail, transaction_return = self._run_transaction(transaction)
        

    #     if passfail:
    #         swdLogger.debug(f'read ctrlstat: {transaction_return["ctrlstat"]["data"]:08x}')
    #     else:
    #         swdLogger.error(f"Error reading core status!")

    #     return passfail, transaction_return


    #AEON specific!

    # #in lua voyII special functions: sWD_stream_memory_Intensity_mode
    # def SWD_bulk_read_old(
    #     self,
    #     bank: int,
    #     address: int,
    #     length: int
    #     ) -> Tuple[bool, List[int]]:

    #     swdLogger.debug(f"bulk read of length {length} from address {address}")

    #     addresses = range(address, address+length, 4)

    #     transaction = {     "name": "bulk_read",
    #                         "namings": None,
    #                         "pattern_src":{
    #                             "steps":       
    #                                   [ {   "template": template_header   } ]
    #                                 + [ {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat"},
    #                                     {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E}, 
    #                                     {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank}, 
    #                                     {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":0x23000012}, 
    #                                     {   "template": template_data_from_host, "packet_request": 0xD1, "write_data":address}, 
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"dummy_0_a"},
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"dummy_0_b"},
    #                                     {   "template": template_data_from_host, "packet_request": 0xD1, "write_data":addresses[0]}, #first address
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"dummy_{addresses[0]}_a"},
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"dummy_{addresses[0]}_b"}, #needs double dummy read
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"value_{addresses[0]}"} ] 
    #                                 + [ x for y in #flatten list
    #                                   [[{   "template": template_data_from_host, "packet_request": 0xD1, "write_data":address_cnt},
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"dummy_{address_cnt}_a"}, #all others only need one dummy read
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"value_{address_cnt}"} ] for address_cnt in addresses[1:]]
    #                                     for x in y] #flatten list
    #                                 + [ {   "template": template_trailer }]}}

    #     passfail, transaction_return = self._run_transaction(transaction)

    #     read_memory = []

    #     if passfail: #on pass
    #         read_memory = [transaction_return[f"value_{address_cnt}"]["data"] for address_cnt in addresses]
        
    #     return passfail, read_memory

    # #in lua voyII special functions: swd_load_FW_to_RAM
    # def SWD_bulk_write_old(
    #     self,
    #     bank: int,
    #     address: int,
    #     data: List[int]
    #     ) -> bool:

    #     transaction = {     "name": "bulk_write",
    #                         "namings": None,
    #                         "pattern_src":{
    #                             "steps":       
    #                                   [ {   "template": template_header   } ]
    #                                 + [ {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat_pre_write"},
    #                                     {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E}, 
    #                                     {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank}, 
    #                                     {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":0x23000012} ]

    #                                 + [ x for y in #flatten list
    #                                   [[{   "template": template_data_from_host, "packet_request": 0xD1, "write_data": address + word_no*4},
    #                                     {   "template": template_data_from_host, "packet_request": 0xDD, "write_data": word},
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":f"dummy_{word_no}"} ] for word_no, word in enumerate(data)]
    #                                     for x in y] #flatten list
    #                                 + [ {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrlstat_post_write"},
    #                                     {   "template": template_trailer }]}}

    #     passfail, transaction_return = self._run_transaction(transaction)

    #     return passfail

    # #in lua (code functions / hw / SWD_interface.lua)
    # def SWD_write_advanced(
    #     self,
    #     address: int,
    #     data: int,
    #     bank: int = 0,
    #     csw_format: int = 0x23000012
    #     ) -> Tuple[bool, Dict]:

    #     transaction = {     "name": "write_advanced",
    #                         "namings": None,
    #                         "pattern_src":{
    #                             "steps":       
    #                                   [ {   "template": template_header   } ]
    #                                 + [ {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E  }, #send abort request
    #                                     {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank  }, #select bank
    #                                     {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":csw_format}, #set csw format
    #                                     {   "template": template_data_from_host, "packet_request": 0xD1, "write_data":address }, #set address
    #                                     {   "template": template_data_from_host, "packet_request": 0xDD, "write_data":data  }, #write data to drw
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"drw"}, #read back drw
    #                                     {   "template": template_data_from_target, "packet_request": 0xBD, "read_data_name":"rdbuff"}, #read back rdbuff
    #                                     {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrl_stat"}, #read back ctrl_stat
    #                                     {   "template": template_trailer }]}}

    #     passfail, transaction_return = self._run_transaction(transaction)

    #     return passfail, transaction_return

    # def SWD_read_advanced(
    #     self,
    #     address,
    #     bank = 0,
    #     csw_format = 0x23000012,
    #     ) -> Tuple[bool, int, Dict]:

    #     address = integerize(address)
    #     bank = integerize(bank)
    #     csw_format = integerize(csw_format)
        
    #     transaction = {     "name": "read_advanced",
    #                         "namings": None,
    #                         "pattern_src":{
    #                             "steps":       
    #                                   [ {   "template": template_header   } ]
    #                                 + [ {   "template": template_data_from_host, "packet_request": 0x81, "write_data":0x0000001E  }, #send abort request
    #                                     {   "template": template_data_from_host, "packet_request": 0x8D, "write_data":bank  }, #select bank
    #                                     {   "template": template_data_from_host, "packet_request": 0xC5, "write_data":csw_format}, #set csw format
    #                                     {   "template": template_data_from_host, "packet_request": 0xD1, "write_data":address }, #set address
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"drw"}, #read back drw
    #                                     {   "template": template_data_from_target, "packet_request": 0xF9, "read_data_name":"drw"}, #read back drw
    #                                     {   "template": template_data_from_target, "packet_request": 0xBD, "read_data_name":"rdbuff"}, #read back rdbuff -> contains the value of interest
    #                                     {   "template": template_data_from_target, "packet_request": 0xB1, "read_data_name":"ctrl_stat"}, #read back ctrl_stat
    #                                     {   "template": template_trailer }]}}

    #     passfail, transaction_return = self._run_transaction(transaction)
    #     if passfail:
    #         return passfail, transaction_return["rdbuff"]["data"], transaction_return
    #     else:
    #         return False, 0

    #in lua voyII special functions
    # def SWD_set_general_purpose_register(
    #     self,
    #     bank,
    #     register_number,
    #     value
    #     ) -> bool :

    #     bank = integerize(bank)
    #     register_number = integerize(register_number)
    #     value = integerize(value)

    #     swdLogger.debug(f"setting general purpose register {register_number} in bank {bank} to {value}")

    #     self.SWD_write_advanced(0xE000EDF8, value, bank)
    #     self.SWD_write_advanced(0xE000EDF4, register_number | 0x00010000, bank)
    #     return self.SWD_read_core_status()[0]

    # def SWD_read_general_purpose_register(
    #     self,
    #     bank: int,
    #     register_number: int
    #     ) -> Tuple[bool, int]:

        
    #     self.SWD_write_advanced(0xE000EDF4, register_number | 0x00010000, bank)
    #     passfail, read_value, transaction_return = self.SWD_read_advanced(0xE000EDF8, bank)

    #     swdLogger.debug(f"read general purpose register {register_number} in bank {bank}, value: {read_value}")

    #     return passfail, read_value

    # def SWD_read_DHCSR(
    #     self,
    #     bank:int
    #     ) -> Tuple[bool, int] :

    #     swdLogger.debug(f"reading DHCSR on bank {bank}")
    #     read_return = self.SWD_read_advanced(0xE000EDF0, bank, 0x23000012)
    #     swdLogger.debug(f"DHCSR: {read_return[1]}")

    #     return read_return[0], read_return[1]

    # def SWD_set_vector_table_address(
    #     self,
    #     start_address: int, 
    #     bank:int
    #     ) -> Tuple[bool, Dict] :
    #     return self.SWD_write_advanced(0xE000ED08 , start_address , bank)

    # def SWD_halt_core(
    #     self,
    #     bank:int
    #     ) -> Tuple[bool, int] :

    #     swdLogger.debug(f"halting core {bank}")
    #     self.SWD_write_advanced(0xE000EDF0 , 0xA05F0003 , bank , 0x23000012)
    #     self.SWD_write_advanced(0xE000EDF0 , 0xA05F0003 , bank , 0x23000012)
    #     read_return = self.SWD_read_advanced(0xE000EDF0, bank, 0x23000012)
    #     swdLogger.debug(f"halt status: {read_return[1]}")
    #     return read_return[0], read_return[1]
    
    # def SWD_release_core(
    #     self,
    #     bank:int
    #     ) -> Tuple[bool, int] :
    #     swdLogger.debug(f"releasing core {bank}")
    #     self.SWD_write_advanced(0xE000EDF0 , 0xA05F0000 , bank , 0x23000012)
    #     self.SWD_write_advanced(0xE000EDF0 , 0xA05F0000 , bank , 0x23000012)
    #     read_return = self.SWD_read_advanced(0xE000EDF0, bank, 0x23000012)
    #     swdLogger.debug(f"halt status: {read_return[1]}")
    #     return read_return[0], read_return[1]

    # def SWD_reset_core(
    #     self,
    #     bank:int
    #     ) -> Tuple[bool, int] :

    #     swdLogger.debug(f"resetting core {bank}")
    #     self.SWD_write_advanced(0xE000EDF0 , 0x05FA0003 , bank , 0x23000012)
    #     self.SWD_write_advanced(0xE000EDFC , 0x01000001 , bank , 0x23000012)
    #     self.SWD_write_advanced(0xE000ED0C , 0x05FA0005 , bank , 0x23000012)
    #     swdLogger.debug("core is reset")
    #     return self.SWD_read_DHCSR(bank)

 

    # def SWD_load_FW_to_RAM(
    #     self,
    #     start_address: int,
    #     fw_path: str,
    #     bank: int) -> bool:

    #     swdLogger.debug(f"loading firmware from {fw_path} to bank {bank} at start address {start_address:08x}")

    #     #read firmware from a file to array
    #     fw_file = open(fw_path, 'r')
    #     fw_data = [int(line, 16) for line in fw_file.readlines()]

    #     self.SWD_halt_core(bank)

    #     self.SWD_bulk_write(bank, start_address, fw_data)

    #     self.SWD_reset_core()

    #     r16_address = 0x01000000
    #     sp_address = fw_data[0]
    #     pc_address = fw_data[1]

    #     self.SWD_set_general_purpose_register(bank, 16, r16_address)
    #     self.SWD_set_general_purpose_register(bank, 15, pc_address)
    #     self.SWD_set_general_purpose_register(bank, 13, sp_address)
    #     self.SWD_set_vector_table_address(start_address, bank)
    #     self.SWD_read_general_purpose_register(bank, 16)
    #     self.SWD_read_general_purpose_register(bank, 15)
    #     self.SWD_read_general_purpose_register(bank, 13)

    #     self.SWD_read_DHCSR(bank)
    #     return self.SWD_release_core(bank) [0]
'''
