from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Tuple


class ams_SWD_communication(ABC):
    """ams_SWD_communication 
    This interface defines the functionality of SWD communication
    """
    class ams_SWD_AP_req_type(Enum):
        """
        This Enum defines the request types used by SWD AP
        """
        DRW = "0x0C"
        TAR = "0x04"
        CSW = "0x00"
        IDR = "0xFC"


    class ams_SWD_DP_req_type(Enum):
        """
        This Enum defines the request types used by SWD DP
        """
        R_IDCODE = "0x00"
        R_CTRL_STAT = "0x04"
        R_RESEND = "0x08"
        R_RDBUFF = "0x0C"
        W_ABORT = "0x00"
        W_CTRL_STAT = "0x04"
        W_SELECT = "0x08"
        W_NA = "0x0C"


    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: SWD object
        """
        pass

    @abstractmethod
    def SWD_init(self) -> bool:
        """SWD_init
        This Method initializes the SWD driver
        
        Returns:
            Bool: True if the Initialization worked, False otherwise
        """    
        pass

    @abstractmethod
    def SWD_configure(
        self, 
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
        ):
        """SWD_configure 
        This method configures the SWD communication.
        If this method is not called the instance uses the defaults given in the HW config file

        Args:
            speed (int): speed of the communication
            bus_voltage (float): bus voltage level in V
            settings_dict (dict): defines any setting that may be implementation specific as interface pins in case of PXI pattern instrument. Defaults to {}.
        """
        pass


    @abstractmethod
    def SWD_pre_init_pattern(self, data: list):
        """SWD_pre_init_pattern 
        This method gives access to the bus and any initial pattern can be sent

        Args:
            data (list): string of characters describing a bitstream with hex values like: 5575AC127E
                there is no maximum length defined for this argument
        """
        pass


    @abstractmethod  
    def SWD_init_sequence(self):
        """
        This method sends the standard initialization pattern
        """
        pass


    @abstractmethod
    def SWD_send_frame(
        self, 
        APnDP: int, 
        RnW: int, 
        address2_3: str, 
        payload: str=''
        ) -> str:
        """SWD_send_frame 
        This method builds SWD frame based on interface spec by standard parameters

        Args:
            APnDP (int): A single bit, indicating whether the Debug Port or the Access Port Access Register is to be accessed. 1 for accessing AP
            RnW (int): A single bit, this bit is 0 for an write access, or 1 for a read access.
            address2_3 (str): Two bits, giving the A[3:2] address field for the DP or AP register address.
            payload (str): payload of the request - if it is not given then reading is going to be done

        Returns:
            str: returns the bytes received from the bus in case of reading
        """
        pass


    @abstractmethod
    def SWD_send_DP_request(
        self, 
        req_type: ams_SWD_DP_req_type=ams_SWD_DP_req_type.R_IDCODE, 
        payload: str=''
        ) -> str:
        """SWD_send_DP_request 
        This method sends DP request
        If payload is not defined at the call, read requiest is going to be sent

        Args:
            req_type (ams_SWD_DP_req_type, optional): type of the request. Defaults to ams_SWD_DP_req_type.R_IDCODE.
            payload (str, optional): payload of the request, in case of reading this field should not be given. Defaults to ''.

        Returns:
            str: returns the bytes received from the bus in case of read request
        """
        pass


    @abstractmethod
    def SWD_send_AP_request(
        self, 
        req_type: ams_SWD_AP_req_type=ams_SWD_AP_req_type.IDR, 
        payload: str=''
        ) -> str:
        """SWD_send_AP_request 
        This method sends DP request
        If payload is not defined at the call, read requiest is going to be sent

        Args:
            req_type (ams_SWD_AP_req_type, optional): type of the request. Defaults to ams_SWD_AP_req_type.IDR.
            payload (str, optional): payload of the request, in case of reading this field should not be given. Defaults to ''.

        Returns:
            str: returns the bytes received from the bus in case of read request
        """
        pass


    @abstractmethod
    def SWD_write_register(
        self, 
        address: str, 
        data: str
        ) -> bool:
        """SWD_write_register 
        Implements the register writing procedure via SWD interface

        Args:
            address (str): memory address to be written
            data (str): memory content to be sent

        Returns:
            bool: returns true if the task is done without any problem
        """
        pass


    @abstractmethod
    def SWD_read_register(
        self, 
        address: str
        ) -> str:
        """SWD_read_register 
        Implements the register reading procedure via SWD interface

        Args:
            address (str): memory address to be red

        Returns:
            str: content of the addressed memory
        """
        pass

    @abstractmethod
    def SWD_parity(
        self, 
        data: str
        ) -> bool:
        """
        This method is responsible to calculate the parity bit

        Return:
            bool: parity bit value
        """
        pass


class ams_I2C_communication(ABC):
    """
    This intreface defines the functions of I2C communnication
    """
    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: I2C communication object
        """
        pass

    @abstractmethod
    def I2C_init(self, device_name):
        """
        This method initializes the I2C communication
        """
        pass


    @abstractmethod
    def I2C_configure(
        self,
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
    ):
        """I2C_configure 
        This method configures the communication

        Args:
            speed (int): communication speed
            bus_voltage (float): bus voltage in V
            settings_dict (dict, optional): defines any implementation setting that may be necessary. Defaults to {}.
        """
        pass


    @abstractmethod
    def I2C_write(
        self, 
        address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_write 
        This method send I2C message in writing mode

        Args:
            address (int): 8 bit address of the IC (7 bit address << 1)
            payload (List[int]): A list of the payload bytes to be transmitted

        Returns:
            bool: True on success, False if the Slave doesn't respond
        """
        pass


    @abstractmethod
    def I2C_read(
        self, 
        address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_read 
        This method send I2C message in reading mode

        Args:
            address (int): 8 bit address of the slave (7 bit address << 1)
            read_count (int): number of bytes to read

        Returns:
            Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """
        pass


    @abstractmethod
    def I2C_register_write(
        self, 
        device_address: int, 
        register_address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_register_write 
        Write bytes to the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to write to
            payload (List[int]): List of bytes to write to the device

        Returns:
            bool: True on success
        """
        pass


    @abstractmethod
    def I2C_register_read(
        self, 
        device_address: int, 
        register_address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_register_read 
        Read bytes from the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to read
            read_count (int): Number of bytes to read

        Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """
        pass



class ams_SPI_communication(ABC):
    """
    This interface defines the function of SPI communication
    """
    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: SPI object
        """
        pass

    @abstractmethod
    def SPI_init(self) -> bool:
        """SPI_init
        This Method initializes the SPI driver
        
        Returns:
            Bool: True if the Initialization worked, False otherwise
        """    
        pass

    @abstractmethod
    def SPI_configure(
        self, 
        speed: int,
        bus_voltage: float,
        settings_dict: dict={}
        ):
        """SPI_configure 
        This method configures the SPI communication.
        If this method is not called the instance uses the defaults given in the HW config file

        Args:
            speed (int): speed of the communication
            bus_voltage (float): bus voltage level in V
            settings_dict (dict): defines any setting that may be implementation specific as interface pins in case of PXI pattern instrument. Defaults to {}.
        """
        pass

    @abstractmethod
    def SPI_transfer(self, payload:str = None, length:int = None) -> Tuple[bool, str]:
        """SPI_transfer
        This method executes the bidirectional data transfer defined by the SPI protocol.
        I can take either the payload and/or the payload length as parameter
        when both are given, the payload is transmitted up to the length. Too short payloads are padded with 0x00
        when no payload length is given, as many bytes as there are in the payload data are transmitted.
        when no payload is given, 0x00 is transmitted for every byte

        Args:
            payload (str, optional): The payload in hexadecimal format. Defaults to None.
            length (int, optional): The number of bytes to be transmitted. Defaults to None.

        Returns:
            Tuple[bool, str]: [True on success, Payload in hex]
        """
        pass